@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -72,11 +72,10 @@ void jed_make_blocal_var (char *name) /*{{{*/
    if ((table == NULL)
        || (table->num == MAX_BLOCAL_VARS_PER_TABLE))
      {
-	table = (Jed_BLocal_Table_Type *) SLmalloc (sizeof (Jed_BLocal_Table_Type));
+	table = (Jed_BLocal_Table_Type *) jed_malloc0 (sizeof (Jed_BLocal_Table_Type));
 	if (table == NULL) 
 	  return;
 
-	memset ((char *) table, 0, sizeof (Jed_BLocal_Table_Type));
 	table->next = CBuf->blocal_table;
 	CBuf->blocal_table = table;
      }
