@@ -59,8 +59,8 @@ int Blink_Flag = 1;
 int Indented_Text_Mode = 0;	       /* if non zero, intent line after wrap */
 int Kill_Line_Feature = 1;	       /* non-zero means kill through eol if bolp  */
 int Jed_Tab_Default = 8;
+int Jed_Wrap_Default = 72;
 int Jed_Case_Search_Default = 0;
-int Jed_Wrap_Column = 72;
 int Jed_Suspension_Not_Allowed = 0;
 int Jed_Use_Tabs = 1;
 
@@ -407,7 +407,7 @@ static int execute_is_ok_hook (SLang_Name_Type *hook)
 int ins_char_cmd (void)
 {
    unsigned char ch;
-   int wrap = Jed_Wrap_Column;
+   int wrap = Buffer_Local.wrap_column;
    int do_blink;
    int did_abbrev = 0;
    SLang_Name_Type *wrapok_hook;
@@ -467,7 +467,7 @@ int ins_char_cmd (void)
 	    && (-1 == jed_insert_byte (ch)))
 	  return -1;
 
-	if (-1 == wrap_line(0))	       /* do not format--- just wrap */
+	if (1 != wrap_line(0))	       /* do not format--- just wrap */
 	  return -1;		       /* line isn't wrapable */
 
 	/* There is a bug involving wrapping a very long line containing
@@ -475,8 +475,8 @@ int ins_char_cmd (void)
 	 * arounds the bug.
 	 */
 	if ((this_line_num == LineNum)
-	    && (ch == ' ')
-	    && (calculate_column () > wrap))
+	    && (ch == ' '))
+	    /* && (calculate_column () > wrap)) */
 	  {
 	     if (0 == jed_right (1))
 	       newline ();
@@ -505,12 +505,15 @@ int ins_char_cmd (void)
 int quoted_insert()
 {
    SLwchar_Type ch;
+   int ins_byte = 1;
 
    CHECK_READ_ONLY
-     if (*Error_Buffer || SLKeyBoard_Quit) return(0);
+   if (*Error_Buffer || SLKeyBoard_Quit) return(0);
+
    if (Repeat_Factor != NULL)
      {
 	ch = *Repeat_Factor;
+	ins_byte = 0;
 	Repeat_Factor = NULL;
      }
    else
@@ -531,10 +534,19 @@ int quoted_insert()
      }
    
    SLKeyBoard_Quit = 0;
-   if (-1 == jed_insert_wchar_n_times(ch, 1))
-     return -1;
-   
-   
+
+   if (ins_byte == 0)
+     {
+	if (-1 == jed_insert_wchar_n_times(ch, 1))
+	  return -1;
+     }
+   else 
+     {
+	unsigned char byte = (unsigned char) ch;
+	if (-1 == jed_insert_nbytes (&byte, 1))
+	  return -1;
+     }
+
    if ((CBuf->syntax_table != NULL)
        && (CBuf->syntax_table->char_syntax[(unsigned char) ch] & CLOSE_DELIM_SYNTAX)
        && !input_pending(&Number_Zero)) blink_match (); /* (ch); */
@@ -700,13 +712,13 @@ static void scroll_completion (int dir)
      {
 	pop_to_buffer (Completion_Buffer);
    	if (dir > 0) pagedown_cmd (); else pageup_cmd ();
-	while (!IS_MINIBUFFER) other_window ();
+	while (!IN_MINI_WINDOW) other_window ();
      }
    else
      {
 	w = JWindow;
 	other_window();
-	if (!IS_MINIBUFFER)
+	if (!IN_MINI_WINDOW)
 	  {
 	     if (dir > 0) pagedown_cmd (); else pageup_cmd ();
 	  }
@@ -743,7 +755,7 @@ int pagedown_cmd()
    int n;
    
    Cursor_Motion = -1;
-   if (IS_MINIBUFFER)
+   if (IN_MINI_WINDOW)
      {
 	scroll_completion (1);
 	return 1;
@@ -803,7 +815,7 @@ int pageup_cmd (void)
    
    Cursor_Motion = -1;
    
-   if (IS_MINIBUFFER)
+   if (IN_MINI_WINDOW)
      {
 	scroll_completion (-1);
 	return 1;
@@ -973,8 +985,8 @@ int indent_line ()
    
    /* CHECK_READ_ONLY */
      
-     if (CLine == CBuf->beg) return(0);
-   
+   if (CLine == CBuf->beg) return(0);
+
    push_spot();
    CLine = CLine->prev;
    get_current_indent (&n);
