@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -13,9 +13,6 @@
 #ifdef __WIN32__
 /* This needs to go first before stdio is included. */
 # include <windows.h>
-# if !defined(__MINGW32__) && !defined(__CYGWIN32__)
-#  define sleep Sleep
-# endif
 # include <fcntl.h>
 # include <io.h>
 # include <sys/stat.h>
@@ -87,16 +84,25 @@ extern unsigned int sleep (unsigned int);
 #endif
 
 #include <errno.h>
+#include <signal.h>
+
+#ifdef VMS
+# include <stat.h>
+# include <rms.h>
+#endif
 
 /* Was anything missed? */
 #ifndef O_RDONLY
-#ifdef VMS
-#include <file.h>
-#else
-#include <fcntl.h>
-#endif
+# ifdef VMS
+#  include <file.h>
+# else
+#  include <fcntl.h>
+# endif
 #endif
 
+#ifndef O_BINARY
+# define O_BINARY	0
+#endif
 
 /*}}}*/
 /*{{{ local inclue files */
@@ -108,6 +114,10 @@ extern unsigned int sleep (unsigned int);
 #include "paste.h"
 #include "ins.h"
 #include "ledit.h"
+#include "userinfo.h"
+#include "hooks.h"
+#include "screen.h"
+#include "sig.h"
 
 /*}}}*/
 
@@ -123,8 +133,6 @@ int Jed_Backup_By_Copying = 0;
 /*{{{ vms_stupid_open */
 
 static int vms_max_rec_size;
-#include <stat.h>
-#include <rms.h>
 static int VMS_write_rfm_fixed;
 int vms_stupid_open(char *file)
 {
@@ -191,75 +199,98 @@ int vms_stupid_open(char *file)
 
 int Require_Final_Newline = 0;
 
+#define _JED_OPEN_READ			0
+#define _JED_OPEN_WRITE			1
+#define _JED_OPEN_APPEND		2
+#define _JED_OPEN_CREATE_EXCL		3
+
 /* 0 = read, 1 = write , 2 = append... */
-static int sys_open(char *file, int acces) /*{{{*/
+static int sys_open (char *file, int acces) /*{{{*/
 {
-   int fp = -1;
    int flags;
-   unsigned int mode;
+   unsigned int perms;
 #ifdef VMS
    char *p, neew[JED_MAX_PATH_LEN];
 #endif
-   
-#ifdef O_BINARY
-	mode = O_BINARY;
-#else
-	mode = 0;
-#endif
 
-   flags = file_status(file);
-   if ((flags < 0) || (flags > 1)) return(fp);
+   flags = file_status (file);
+   if ((flags < 0) || (flags > 1)) 
+     return -1;			       /* directory? */
    
-   /* on VMS I cheat since I do not want to deal with RMS at this point */
 #ifdef VMS
+   /* on VMS I cheat since I do not want to deal with RMS at this point */
    VMS_write_rfm_fixed = 0;
-   strcpy(neew, file);
+   safe_strcpy(neew, file, sizeof (neew));
    p = neew; while (*p) if (*p == ';') *p = 0; else p++;
    
-   if (acces == 0) fp = open(file, O_RDONLY, "ctx=rec","mbf=8","mbc=32","rop=RAH","shr=upi,get,put");
-   else if (acces == 1)
+   switch (acces)
      {
-	fp = vms_stupid_open(neew);
+      case _JED_OPEN_READ:
+	return open(file, O_RDONLY, "ctx=rec","mbf=8","mbc=32","rop=RAH","shr=upi,get,put");
+
+      case _JED_OPEN_WRITE:
+      case _JED_OPEN_CREATE_EXCL:      /* FIXME */
+	return vms_stupid_open (neew);
+
+      case _JED_OPEN_APPEND:
+	return open (file, O_WRONLY | O_APPEND | O_CREAT | O_BINARY);
+      default:
+	return -1;
      }
-   
-   else if (acces == 2) fp = open(file, O_WRONLY | O_APPEND | O_CREAT | mode);
+
 #else
-   
-   switch(acces)
+
+# ifdef IBMPC_SYSTEM
+   perms = S_IREAD | S_IWRITE;
+# else
+   perms = 0666;
+# endif
+
+   switch (acces)
      {
-      case 0: flags =  O_RDONLY; 
+      case _JED_OPEN_READ:
+	flags = O_RDONLY; 
 	break;
-      case 1: flags =  O_WRONLY | O_CREAT | O_TRUNC;
+      case _JED_OPEN_WRITE: 
+	flags = O_WRONLY | O_CREAT | O_TRUNC;
 	break;
-      case 2: flags =  O_WRONLY | O_CREAT | O_APPEND; 
+      case _JED_OPEN_APPEND: 
+	flags = O_WRONLY | O_CREAT | O_APPEND; 
 	break;
-      default: return(fp);
-     }
-   
-   flags |= mode;
-   
-#if defined (IBMPC_SYSTEM)
-   fp = open(file, flags, S_IREAD | S_IWRITE); 
-#else
-   fp = open(file, flags, 0666); 
+      case _JED_OPEN_CREATE_EXCL:
+	flags = O_WRONLY | O_CREAT | O_EXCL;
+#ifndef IBMPC_SYSTEM
+	perms = 0600;
 #endif
+	break;
+	
+      default:
+	return -1;
+     }
+   
+   flags |= O_BINARY;
+   
+   return open(file, flags, perms);
 #endif /* VMS */
-   return(fp);
 }
 
 /*}}}*/
 
-char *file_type(char *file) /*{{{*/
+char *file_type (SLFUTURE_CONST char *file) /*{{{*/
 {
    char *p, *psave;
-   if ((file == (char *) NULL) || (*file == 0)) return( (char *) NULL);
+   
+   if ((file == NULL)
+       || (*file == 0))
+     return NULL;
 
    file = extract_file(file);
-   p = file; while (*p != 0) p++;
+   p = (char *) file; while (*p != 0) p++;
    psave = p;
    while((p != file) && (*p != '.')) p--;
    if (*p == '.') p++;
-   if (p == file) return(psave); else return(p);
+   if (p == file) return psave;
+   return p;
 }
 
 /*}}}*/
@@ -267,7 +298,7 @@ char *file_type(char *file) /*{{{*/
 
 void jed_set_buffer_ctime (Buffer *b)
 {
-   char file[JED_MAX_PATH_LEN + 1];
+   char *file;
 
    if ((b->file == NULL)
        || (b->file[0] == 0))
@@ -275,14 +306,101 @@ void jed_set_buffer_ctime (Buffer *b)
 	b->c_time = 0;
 	return;
      }
-   
-   strcpy (file, b->dir);
-   strcat (file, b->file);
 
-   if (0 == sys_file_mod_time (file))
-     b->c_time = 0;
-   else
-     b->c_time = sys_time ();
+   if (NULL == (file = SLang_concat_slstrings (b->dir, b->file)))
+     {
+	b->c_time = 0;
+	return;
+     }
+   b->c_time = sys_file_mod_time (file);
+   
+   SLang_free_slstring (file);
+}
+
+#if (defined(__MSDOS__) || defined(__WIN32__)) && !defined(W_OK)
+# define W_OK 2
+# define F_OK 0
+#endif
+
+#ifdef __GO32__
+# define access i386_access
+#endif
+
+int jed_file_is_readonly (char *file, int respect_perms)
+{
+   int ro = 0;
+   struct stat st;
+
+   if (respect_perms)
+     {
+	/* respect the file's permissions.  If any write permission bit is
+	 * is set, then consider it a writable candidate.
+	 */
+#if defined(S_IWGRP) && defined(S_IWOTH)
+	if (0 == stat (file, &st))
+	  ro = (0 == (st.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)));
+#endif
+     }
+
+#ifdef W_OK
+   if (ro == 0)
+     {
+	if (0 == access(file, F_OK))
+	  {
+	     if (-1 == access(file, W_OK))
+	       ro = 1;
+	  }
+	else
+	  {
+	     /* file does not exist.  Can we write to the directory? */
+	     char *dir;
+	     
+	     if (-1 == jed_dirfile_to_dir_file (file, &dir, NULL))
+	       return -1;
+
+# if defined(IBMPC_SYSTEM)
+	     /* knock of slash since some primitive systems cannot handle the 
+	      * final slash in a path name. 
+	      */
+	       {
+		  unsigned int len = strlen (dir);
+		  if (len > 3)   /* allow C:/, but not C:/xx/ */
+		    dir[len-1] = 0;
+	       }
+# endif
+	     if ((0 == access (dir, F_OK))
+# ifdef VMS
+		 && (-1 == access (dir, X_OK))
+#else
+		 && (-1 == access (dir, W_OK))
+#endif
+		)
+	       ro = 1;
+	     
+	     SLfree (dir);
+	  }
+     }
+#endif				       /* W_OK */
+   return ro;
+}
+
+int jed_buffer_file_is_readonly (Buffer *b)
+{
+   char *file;
+   int ro;
+
+   if ((b->file == NULL)
+       || (b->file[0] == 0))
+     return 0;
+
+   file = jed_dir_file_merge (b->dir, b->file);
+   if (file == NULL)
+     return -1;
+
+   ro = jed_file_is_readonly (file, 1);
+
+   SLfree (file);
+   return ro;
 }
 
 void set_file_modes (void) /*{{{*/
@@ -292,59 +410,33 @@ void set_file_modes (void) /*{{{*/
    if (CBuf == NULL) return;
 
    jed_set_buffer_ctime (CBuf);
+   if (1 == jed_buffer_file_is_readonly (CBuf))
+     CBuf->flags |= READ_ONLY;
+
+   type = NULL;
 
    if ((CBuf->file != NULL)
        && (CBuf->file[0] != 0))
      {
-#if (defined(__MSDOS__) || defined(__WIN32__)) && !defined(W_OK)
-# define W_OK 2
-# define F_OK 0
-#endif
-
-#ifdef W_OK
-#ifdef __GO32__
-# define access i386_access
-#endif
-	char name [JED_MAX_PATH_LEN + 256];
-#ifndef VMS
-	  {
-	     /* COmment:  I need a more generic call to contruct a directory
-	      * name.  site.sl already does this properly.
-	      */
-	     int n;
-	     strcpy (name, CBuf->dir);
-	     n = strlen (name);
-	     if (n >
-#ifdef IBMPC_SYSTEM
-		 3		       /* allow C:/file */
-#else
-		 1		       /* allow /file */
-#endif
-		 )
-	       {
-		  name[n - 1] = 0;     /* knock off slash */
-		  if (!access (name, F_OK) && (access (name, W_OK)))
-		    CBuf->flags |= READ_ONLY;
-	       }
-	  }
-#endif  /* NOT VMS */
-	sprintf (name, "%s%s", CBuf->dir, CBuf->file);
-	if (!access(name, F_OK) && access(name, W_OK))
-	  CBuf->flags |= READ_ONLY;
-#endif /* W_OK */
 	CBuf->flags |= AUTO_SAVE_BUFFER;
 	CBuf->hits = 0;
-	type = file_type(CBuf->file);
+
+	if (type == NULL) 
+	  type = file_type (CBuf->file);
      }
-   else type = NULL;
 
    CBuf->modes = NO_MODE;
 #if JED_HAS_LINE_ATTRIBUTES
    CBuf->min_unparsed_line_num = 1;
    CBuf->max_unparsed_line_num = Max_LineNum + CBuf->nup;
 #endif
+
    if (type != NULL)
-     SLang_run_hooks("mode_hook", 1, type);
+     {
+	if (1 != jed_va_run_hooks ("_jed_set_mode_hooks", 
+				   JED_HOOKS_RUN_UNTIL_NON_0, 1, type))
+	  (void) SLang_run_hooks("mode_hook", 1, type);
+     }
 }
 
 /*}}}*/
@@ -353,20 +445,33 @@ void set_file_modes (void) /*{{{*/
 
 int read_file(char *file) /*{{{*/
 {
-   int fp;
    int n, status;
 
-   if ((fp = sys_open(file, 0)) < 0)
+   status = jed_va_run_hooks ("_jed_read_file_hooks", JED_HOOKS_RUN_UNTIL_NON_0, 1, file);
+   if (status < 0)
+     return -1;
+   
+   if (status > 0)
+     /* FIXME!! Is this true?  What if no lines have been read. */
+     n = Max_LineNum;
+   else
      {
-	status = file_status(file);
-	if (!status) return(-1);  /* file does not exist */
-	return(-2); /* exists but not readable */
+	int fp;
+
+	if ((fp = sys_open(file, _JED_OPEN_READ)) < 0)
+	  {
+	     status = file_status(file);
+	     if (!status) return(-1);  /* file does not exist */
+	     return(-2); /* exists but not readable */
+	  }
+	n = read_file_pointer(fp);
+	close(fp);
      }
 
-   n = read_file_pointer(fp);
-   close(fp);
    eob();
-   if ('\n' == *(CLine->data + Point)) make_line(2);
+   if ((Point < CLine->len)
+       && ('\n' == *(CLine->data + Point)))
+     make_line(2);
 
    VFile_Mode = VFILE_TEXT;
    return n;
@@ -383,8 +488,11 @@ int insert_file_pointer(VFILE *vp) /*{{{*/
    while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
      {
 	n++;
-	if (SLang_Error) break;
-	quick_insert(vbuf, (int) num);
+	if (SLang_get_error ())
+	  break;
+
+	if (-1 == jed_quick_insert (vbuf, (int) num))
+	  return -1;
      }
    return(n);
 }
@@ -394,8 +502,13 @@ int insert_file(char *file) /*{{{*/
 {
    VFILE *vp;
    int n;
-   
-   if (NULL == (vp = vopen(file, 0, VFile_Mode))) return(-1);
+   unsigned int un;
+
+   un = Max_LineNum;
+   if (1 == jed_va_run_hooks ("_jed_insert_file_hooks", JED_HOOKS_RUN_UNTIL_NON_0, 1, file))
+     return (int) (Max_LineNum - un);
+
+   if (NULL == (vp = vopen(file, 0, VFile_Mode))) return -1;
    n = insert_file_pointer(vp);
    vclose(vp);
    return(n);
@@ -435,6 +548,7 @@ static int jed_write1(int fd, char *b, unsigned int n) /*{{{*/
    
    while (total < n)
      {
+	int dlen;
 	len = n - total;
 #ifdef VMS
 	if (VMS_write_rfm_fixed)
@@ -449,17 +563,23 @@ static int jed_write1(int fd, char *b, unsigned int n) /*{{{*/
 	if (bmax == b) bmax = b + (len - 1); /* cannot be helped */
 	len = (int) (bmax - b) + 1;
 #endif
-	while (-1 == (len = write (fd, b, len)))
+	while (-1 == (dlen = write (fd, b, len)))
 	  {
 #ifdef EINTR
 	     if (errno == EINTR)
-	       continue;
+	       {
+		  if (0 == jed_handle_interrupt ())
+		    continue;
+	       }
 #endif
 #ifdef EAGAIN
 	     if (errno == EAGAIN)
 	       {
-		  sleep (1);
-		  continue;
+		  if (0 == jed_handle_interrupt ())
+		    {
+		       jed_sleep (1);
+		       continue;
+		    }
 	       }
 #endif
 #ifdef ENOSPC
@@ -469,12 +589,12 @@ static int jed_write1(int fd, char *b, unsigned int n) /*{{{*/
 		  return total;
 	       }
 #endif
-	     msg_error ("Write Failed: Unknown Error.");
+	     jed_verror ("Write Failed: (%d bytes attemped, errno = %d)", len, errno);
 	     return total;
 	  }
 	
-	total += (unsigned int) len;
-	b += len;
+	total += (unsigned int) dlen;
+	b += dlen;
      }
    return total;
 #else
@@ -606,7 +726,7 @@ static int jed_write(int fd, char *b, unsigned int n) /*{{{*/
 /*}}}*/
 
 /* returns -1 on failure */
-int write_region_to_fp(int fp) /*{{{*/
+int write_region_to_fp (int fp) /*{{{*/
 {
    register int pnt, len;
    register Line *first, *last;
@@ -632,7 +752,7 @@ int write_region_to_fp(int fp) /*{{{*/
    if (!check_region(&Number_One)) return(-1);
    last = CLine; last_pnt = Point;
 
-   pop_mark(&Number_One);
+   jed_pop_mark(1);
    first = CLine; pnt = Point;
 
    /* first should never be null without hitting last first.  If this
@@ -648,12 +768,12 @@ int write_region_to_fp(int fp) /*{{{*/
 	/* This goes here inside the loop because it is possible for external
 	   events to set error_buffer */
 	pnt = 0;
-	if (SLang_Error) break;
+	if (SLang_get_error ()) break;
 	first = first->next;
 	n++;
      }
 
-   if (!SLang_Error && (last_pnt != 0))
+   if (!SLang_get_error () && (last_pnt != 0))
      {
 	len = last_pnt - pnt;
 	if (len != jed_write(fp, (char *) (last->data + pnt), len))
@@ -673,7 +793,7 @@ int write_region_to_fp(int fp) /*{{{*/
    /* Now flush output buffer if necessary */
    
    len = (int) (Output_Bufferp - Output_Buffer);
-   if (!SLang_Error && len) if (len != jed_write1(fp, Output_Buffer, len))
+   if (!SLang_get_error () && len) if (len != jed_write1(fp, Output_Buffer, len))
      {
 	msg_error(err);
      }
@@ -683,7 +803,7 @@ int write_region_to_fp(int fp) /*{{{*/
    
    pop_spot();
    VFile_Mode = VFILE_TEXT;
-   if (SLang_Error) return(-1);
+   if (SLang_get_error ()) return(-1);
    return(n);
 }
 
@@ -697,9 +817,12 @@ static int jed_close (int fp) /*{{{*/
 #ifndef IBMPC_SYSTEM
 	if (errno == EINTR) 
 	  {
-	     errno = 0;
-	     sleep (1);
-	     continue;
+	     if (-1 == jed_handle_interrupt ())
+	       {
+		  errno = 0;
+		  jed_sleep (1);
+		  continue;
+	       }
 	  }
 #endif
 #endif
@@ -711,31 +834,73 @@ static int jed_close (int fp) /*{{{*/
 
 /*}}}*/
 
-int write_region (char *file) /*{{{*/
+static int write_region_internal (char *file, int omode) /*{{{*/
 {
-   int fp;
+   int fd;
    int n;
+   unsigned int num_lines;
 
-   if (!check_region(&Number_Zero)) return(-1);
-   if ((fp = sys_open(file, 1)) < 0)
+#if JED_HAS_EMACS_LOCKING
+   if (-1 == jed_lock_file (file))
+     return -1;
+#endif
+   if (!check_region(&Number_Zero)) 
+     return -1;
+   
+   num_lines = jed_count_lines_in_region ();
+
+   switch (omode)
      {
-	jed_verror ("Unable to open %s for writing.", file);
-	return -1;
+      case _JED_OPEN_WRITE:
+      case _JED_OPEN_CREATE_EXCL:
+	n = jed_va_run_hooks ("_jed_write_region_hooks",
+			      JED_HOOKS_RUN_UNTIL_NON_0, 1, file);
+	break;
+	
+      default:
+	n = jed_va_run_hooks ("_jed_append_region_hooks",
+			      JED_HOOKS_RUN_UNTIL_NON_0, 1, file);
+	break;
      }
-   n = write_region_to_fp(fp);
-   if (-1 == jed_close (fp)) n = -1;
 
-   return(n);
+   if (n > 0)
+     n = num_lines;
+   else if (n < 0)
+     n = -1;
+   else if (n == 0)
+     {
+	if ((fd = sys_open(file, omode)) < 0)
+	  {
+	     jed_verror ("Unable to open %s for writing.", file);
+	     return -1;
+	  }
+
+	n = write_region_to_fp (fd);
+	if (n == -1)
+	  (void) jed_close (fd);
+	else if (-1 == jed_close (fd))
+	  n = -1;
+     }
+   
+#if JED_HAS_EMACS_LOCKING
+   if (n != -1) jed_unlock_file (file);
+#endif
+   return n;
 }
 
 /*}}}*/
 
+int write_region (char *file)
+{
+   return write_region_internal (file, _JED_OPEN_WRITE);
+}
+
 
 /* returns -1 on failure and number of lines on success */
-static int write_file(char *file) /*{{{*/
+static int write_file_internal (char *file, int omode) /*{{{*/
 {
    Mark *m;
-   int n = -1;
+   int n;
    int fnl;
       
 #ifdef VMS
@@ -762,7 +927,7 @@ static int write_file(char *file) /*{{{*/
 #endif
    
    bob();
-   push_mark();  m = CBuf->marks;
+   jed_push_mark();  m = CBuf->marks;
    eob();
    fnl = Require_Final_Newline;
    if (CBuf->flags & BINARY_FILE) 
@@ -775,11 +940,12 @@ static int write_file(char *file) /*{{{*/
 #endif
      }
 
-   n = write_region(file);
-   
+
+   n = write_region_internal (file, omode);
+
    Require_Final_Newline = fnl;
    VFile_Mode = VFILE_TEXT;
-   if (m == CBuf->marks) pop_mark(&Number_Zero);
+   if (m == CBuf->marks) jed_pop_mark(0);
    
 #if JED_HAS_SAVE_NARROW
    jed_pop_narrow ();
@@ -793,21 +959,18 @@ static int write_file(char *file) /*{{{*/
 
 int append_to_file(char *file) /*{{{*/
 {
-   int fp;
-   int n;
-
-   if ((fp = sys_open(file, 2)) < 0) return(-1);
-   n = write_region_to_fp(fp);
-   if (-1 == jed_close (fp)) n = -1;
+   int status;
+   
+   status = write_region_internal (file, _JED_OPEN_APPEND);
    check_buffers();
-   return(n);
+   return status;
 }
 
 /*}}}*/
 
 /*}}}*/
 
-static int make_autosave_filename(char *save, char *dir, char *file) /*{{{*/
+static int make_autosave_filename(char *save, unsigned int buflen, char *dir, char *file) /*{{{*/
 {
    char *s;
 
@@ -817,19 +980,19 @@ static int make_autosave_filename(char *save, char *dir, char *file) /*{{{*/
    if (1 == SLang_run_hooks ("make_autosave_filename", 2, dir, file))
      {
 	if (SLang_pop_slstring(&s)) return(0);
-	strncpy(save, s, JED_MAX_PATH_LEN - 2);
-	save[JED_MAX_PATH_LEN - 1] = 0;
+	strncpy(save, s, buflen);
+	save[buflen - 1] = 0;
 	SLang_free_slstring (s);
      }
    else
      {
 #if defined(IBMPC_SYSTEM)
-	sprintf(save, "%s#%s", dir, file);
+	SLsnprintf (save, buflen, "%s#%s", dir, file);
 #else
 # ifdef VMS
-	sprintf(save, "%s_$%s;1", dir, file);
+	SLsnprintf (save, buflen, "%s_$%s;1", dir, file);
 # else
-	sprintf(save, "%s#%s#", dir, file);
+	SLsnprintf (save, buflen, "%s#%s#", dir, file);
 # endif
 #endif
      }
@@ -841,8 +1004,9 @@ static int make_autosave_filename(char *save, char *dir, char *file) /*{{{*/
 #ifndef VMS
 int jed_copy_file (char *from, char *to) /*{{{*/
 {
-   int mode;
-   short uid, gid;
+   mode_t mode;
+   uid_t uid;
+   gid_t gid;
    FILE *fp0, *fp1;
    char buf[0x7FFF];
    unsigned int readlen;
@@ -861,8 +1025,26 @@ int jed_copy_file (char *from, char *to) /*{{{*/
    
    fp0 = fopen (from, "rb");
    if (fp0 == NULL) return -1;
-   
+
+#ifdef REAL_UNIX_SYSTEM
+   (void) sys_delete_file (to);
+   /* Try to avoid a race condition (code derived from Colin Phipps <crp22@cam.ac.uk>. */
+   ret = sys_open (to, _JED_OPEN_CREATE_EXCL);
+   if (ret == -1)
+     {
+	(void) fclose (fp0);
+	return -1;
+     }
+   if (NULL == (fp1 = fdopen (ret, "wb")))
+     {
+	(void) close (ret);
+	(void) sys_delete_file (to);
+	/* Drop */
+     }
+#else
    fp1 = fopen (to, "wb");
+#endif
+
    if (fp1 == NULL) 
      {
 	(void) fclose (fp0);
@@ -954,13 +1136,14 @@ unsigned long sys_file_mod_time(char *file) /*{{{*/
 
 /*}}}*/
 
-int write_file_with_backup(char *dir, char *file) /*{{{*/
+int write_file_with_backup(char *dirfile) /*{{{*/
 {
-   char neew[JED_MAX_PATH_LEN]; char autosave_file[JED_MAX_PATH_LEN];
+   char autosave_file[JED_MAX_PATH_LEN];
    int n;
-   int mode, do_mode;
-   short uid, gid;
-   char *mode_string;
+   int do_mode;
+   uid_t uid;
+   gid_t gid;
+   mode_t mode;
 #ifndef VMS
    char *old = NULL;
    int backup_went_ok;
@@ -968,33 +1151,47 @@ int write_file_with_backup(char *dir, char *file) /*{{{*/
 #ifdef __os2__
    PHOLDFEA EAs;
 #endif
-   
-   if (*file == 0) return(-1);
+   char *dir, *file;
 
-   sprintf(neew, "%s%s", dir, file);
+   if (NULL == (dirfile = jed_expand_link (dirfile)))
+     return -1;
+
+   if (-1 == jed_dirfile_to_dir_file (dirfile, &dir, &file))
+     {
+	SLfree (dirfile);
+	return -1;
+     }
+
+   if (*file == 0)
+     {
+	SLfree (dir);
+	SLfree (dir);
+	SLfree (dirfile);
+	return -1;
+     }
 
    *autosave_file = 0;
    if (CBuf->flags & AUTO_SAVE_BUFFER)
-      (void) make_autosave_filename(autosave_file, dir, file);
+     (void) make_autosave_filename(autosave_file, sizeof (autosave_file), dir, file);
+
+   do_mode = sys_chmod(dirfile, 0, &mode, &uid, &gid);
+   if ((do_mode != 0) && (do_mode != 1)) 
+     {
+	SLfree (dir);
+	SLfree (file);
+	SLfree (dirfile);
+	return -1;
+     }
 
-   do_mode = sys_chmod(neew, 0, &mode, &uid, &gid);
-   if ((do_mode != 0) && (do_mode != 1)) return -1;
    /* IF do_mode == 0, then the file does not exist.  This means that 
     * there is nothing to backup.  If do_mode == 1, the file is a regular
     * file which we do want to backup.
     */
-   if (NULL == (mode_string = CBuf->mode_string))
-     mode_string = "";
-   if (-1 == SLang_run_hooks ("save_buffer_hook", 2, neew, mode_string))
-     {
-	auto_save_buffer (CBuf);
-	return -1;
-     }
 
 #ifndef VMS
    
 # ifdef __os2__
-   EAs = QueryEAs (neew);
+   EAs = QueryEAs (dirfile);
 # endif
    
    backup_went_ok = 0;
@@ -1005,51 +1202,55 @@ int write_file_with_backup(char *dir, char *file) /*{{{*/
 	if ((0 == SLang_pop_slstring(&old))
 	    && (*old != 0))
 	  {
-	     backup_went_ok = (0 == perform_backup (neew, old, 0));
+	     backup_went_ok = (0 == perform_backup (dirfile, old, 0));
 	  }
      }
 #endif				       /* NOT VMS */
 
-   if (-1 != (n = write_file(neew)))
+   n = write_file_internal (dirfile, _JED_OPEN_WRITE);
+   /* This is for NFS time problems.  Even if the write failed, modify the 
+    * buffer's ctime because otherwise what is on the disk (a partial file) 
+    * will appear newer than the buffer.
+    */
+   CBuf->c_time = sys_file_mod_time(dirfile);
+
+   if (n != -1)
      {
 	if (*autosave_file)
-	  {
-	     sys_delete_file(autosave_file);
-	  }
-	     
+	  (void) sys_delete_file (autosave_file);
+
 	if (do_mode) /* must be an existing file, so preserve mode */
 	  {
 #if defined(__MSDOS__)
 	     /* Want the archive bit set since this is a new version */
 	     mode |= 1 << 5;
 #endif
-	     sys_chmod (neew, 1, &mode, &uid, &gid);
+	     sys_chmod (dirfile, 1, &mode, &uid, &gid);
 #ifdef __os2__
-	     WriteEAs (neew, EAs);
+	     WriteEAs (dirfile, EAs);
 #endif
 	  }
-	/* This is for NFS time problems */
-	CBuf->c_time = sys_file_mod_time(neew);
 	
 	/* Since we wrote the buffer to the file, it is not modified. */
-	if (
-#if JED_FILE_PRESERVE_CASE
-	    (0 == strcmp(file, CBuf->file)) && (0 == strcmp(dir, CBuf->dir))
-#else
-	    (0 == jed_case_strcmp (file, CBuf->file))
-	     && (0 == jed_case_strcmp (dir, CBuf->dir))
-#endif
-	    )
+	if (CBuf == find_file_buffer (dirfile))
 	  CBuf->flags &= ~FILE_MODIFIED;
 	
-	/* CBuf->c_time = sys_time(); */
+	mark_buffer_modified (CBuf, 0, 1);
      }
 #ifndef VMS
    /* error -- put it back */
-   else if (backup_went_ok) perform_backup (old, neew, 1); 
-   
+   else if (backup_went_ok) perform_backup (old, dirfile, 1);
+
    if (old != NULL) SLang_free_slstring (old);
 #endif
+
+#ifdef REAL_UNIX_SYSTEM
+   (void) jed_get_inode_info (dirfile, &CBuf->device, &CBuf->inode);
+#endif
+   SLfree (dir);
+   SLfree (file);
+   SLfree (dirfile);
+
    return(n);
 }
 
@@ -1060,17 +1261,24 @@ void auto_save_buffer(Buffer *b) /*{{{*/
    char tmp[JED_MAX_PATH_LEN];
    Buffer *old_buf;
    unsigned int vfm = VFile_Mode;
-   
+   char *dir, *file;
+
    if (b == NULL) return;
    b->hits = 0;
    if ((0 == (b->flags & BUFFER_MODIFIED))
        || (0 == (b->flags & (AUTO_SAVE_BUFFER | AUTO_SAVE_JUST_SAVE))))
      return;
 
+   if (b->canonical_dirfile == NULL)
+     return;
+
 #if !JED_HAS_SAVE_NARROW
    if (b->narrow != NULL) return;
 #endif
 
+   if (-1 == jed_dirfile_to_dir_file (b->canonical_dirfile, &dir, &file))
+     return;
+
    old_buf = CBuf;
    switch_to_buffer(b);
    
@@ -1079,22 +1287,19 @@ void auto_save_buffer(Buffer *b) /*{{{*/
 
    if (b->flags & AUTO_SAVE_BUFFER)
      {
-	if (make_autosave_filename(tmp, b->dir, b->file))
+	if (make_autosave_filename(tmp, sizeof (tmp), dir, file))
 	  {
 	     flush_message("autosaving..."); 
-	     sys_delete_file(tmp);
-	     write_file(tmp);
-#ifndef IBMPC_SYSTEM
-	     (void) chmod (tmp, 0600);
-#endif
+	     (void) sys_delete_file(tmp);
+	     (void) write_file_internal (tmp, _JED_OPEN_CREATE_EXCL);
 	     message("autosaving...done");
 	  }
       }
-   else if (write_file_with_backup(b->dir, b->file) >= 0)
-     {
-	b->flags &= ~BUFFER_MODIFIED;
-     }
+   else (void) write_file_with_backup(b->canonical_dirfile);
 
+   SLfree (file);
+   SLfree (dir);
+   
    switch_to_buffer(old_buf);
    VFile_Mode = vfm;
 }
@@ -1135,7 +1340,7 @@ static int is_link(char *f, char *f1) /*{{{*/
    l--;
    if ((l > 1) && (f[l] == '/'))
      {
-	strcpy(work, f);
+	safe_strcpy(work, f, sizeof (work));
 	is_dir = 1;
 	f = work;
 	f[l] = 0;
@@ -1178,33 +1383,31 @@ static int is_link(char *f, char *f1) /*{{{*/
 
 /*}}}*/
 
-/* This routine should be modified to work with all components of the path
- * and not just the last component.  For example, suppose /usr/lib/jed is a 
- * link to /usr/local/src/jed and one tries to read the file
- * /usr/lib/jed/lib/site.sl.  This should be expanded to 
- * /usr/local/lib/jed/lib/site.sl.  Currently, it will only have an effect if
- * site.sl is itself a link.
- */
-char *expand_link(char *f) /*{{{*/
+/*}}}*/
+#endif
+
+
+#ifdef REAL_UNIX_SYSTEM
+static char *expand_link_2(char *f) /*{{{*/
 {
    char work[JED_MAX_PATH_LEN], lnk[JED_MAX_PATH_LEN];
-   char *d = expand_filename(f);
- 
-  
-   while (is_link(d, lnk))
+   char *d = jed_standardize_filename (f);
+   unsigned int max_num = 20;
+
+   while ((d != NULL) && is_link(d, lnk))
      {
 	char *w, *w1, ch;
 	
 	if (*lnk == '/')	       /* absolute */
-	  strcpy (work, lnk);
+	  safe_strcpy (work, lnk, sizeof (work));
 	else
 	  {
-	     strcpy (work, d);
+	     safe_strcpy (work, d, sizeof (work));
 	     *(extract_file(work)) = 0;
-	     strcat (work, lnk);
+	     safe_strcat (work, lnk, sizeof (work));
 	  }
-	
-	/* Collapse multiple // since expand_filename will interpret them
+
+	/* Collapse multiple // since jed_standardize_filename_static will interpret them
 	 * differently.
 	 */
 	w = w1 = work;
@@ -1215,19 +1418,172 @@ char *expand_link(char *f) /*{{{*/
 	  }
 	*w = 0;
 	
-	d = expand_filename(work);
+	SLfree (d);
+	max_num--;
+	if (max_num == 0)
+	  {
+	     jed_verror ("possible circular-link detected");
+	     return NULL;
+	  }
+	d = jed_standardize_filename(work);
      }
    
    return (d);
 }
-
 /*}}}*/
+#endif				       /* REAL_UNIX_SYSTEM */
 
-/*}}}*/
+#ifdef REAL_UNIX_SYSTEM
+static char *expand_link_1(char *f) /*{{{*/
+{
+   char *d = SLmake_string (f);
+   
+   if (d == NULL)
+     return d;
+
+   f = d;
+
+   while (1)
+     {
+	char *new_d, *d1;
+	
+	if (*f == '/')
+	  f++;
+
+	while (*f && (*f != '/'))
+	  f++;
+	
+	if (*f == 0)
+	  break;
+	
+	*f++ = 0;
+
+	new_d = expand_link_2 (d);
+	if (new_d == NULL)
+	  {
+	     SLfree (d);
+	     return NULL;
+	  }
+
+	d1 = SLpath_dircat (new_d, f);
+	f = d1 + strlen (new_d);
+	SLfree (d);
+	SLfree (new_d);
+	if (d1 == NULL)
+	  return NULL;
+	d = d1;
+     }
+   
+   /* Now get filename component */
+   f = expand_link_2 (d);
+   SLfree (d);
+   return f;
+}
+#endif				       /* REAL_UNIX_SYSTEM */
+
+/* Malloced string returned */
+char *jed_expand_link (char *f)
+{
+#ifndef REAL_UNIX_SYSTEM
+   return jed_standardize_filename (f);
+#else
+   char *f1, *f2;
+   unsigned int len;
+   
+   /* Look for, e.g., /usr/local//tmp/foo and extract /tmp/foo.
+    * Do this before attempting to expand a link since links could 
+    * contain such constructs where they have a different meaning.
+    * 
+    * Do this only if // occurs because ../ constructs should be preserved
+    * if a subdir is a link to another.  That is consider:
+    * 
+    *    /path/a/b/../c.txt
+    * 
+    * jed_standardize_filename will convert this to
+    *  
+    *    /path/a/c.txt
+    * 
+    * However, if b is a symlink to /tmp/a/b then we really want the file
+    * /tmp/a/c.txt.
+    */
+#if 1   
+   len = strlen (f);
+   if (NULL == (f = SLmake_nstring (f, len)))
+     return NULL;
+   
+   f2 = f + len;
+   while (f2 > f)
+     {
+	if ((*f2-- == '/') && (*f2 == '/'))
+	  {
+	     f2++;
+	     break;
+	  }
+     }
+#else
+   if (NULL == (f = jed_standardize_filename (f)))
+     return NULL;
+   f2 = f;
 #endif
 
-void visit_file(char *dir, char *file) /*{{{*/
+   if (NULL == (f1 = expand_link_1 (f2)))
+     {
+	SLfree (f);
+	return NULL;
+     }
+
+   if (0 == strcmp (f, f1))
+     {
+	SLfree (f);
+	return f1;
+     }
+   SLfree (f);
+
+   f = f1;
+   /* Keep expanding the link until nothing changes */
+   while (1)
+     {
+	if (NULL == (f1 = expand_link_1 (f)))
+	  {
+	     SLfree (f);
+	     return NULL;
+	  }
+
+	if (0 == strcmp (f, f1))
+	  {
+	     SLfree (f1);
+	     return f;
+	  }
+	SLfree (f);
+	f = f1;
+     }
+#endif
+}
+
+
+/*}}}*/
+
+void visit_file (char *dir, char *file) /*{{{*/
 {
+#ifdef VMS
+   char *ff;
+
+   file = SLmake_string (file);
+   if (file == NULL)
+     return;
+
+   ff = file; 
+   while (*ff) 
+     {
+	if (*ff == ';')
+	  {
+	     *ff = 0; 
+	     break;
+	  }
+	ff++;
+     }
+#endif
+
    if (
 #if JED_FILE_PRESERVE_CASE
        strcmp(file, CBuf->file) || strcmp(dir, CBuf->dir)
@@ -1247,27 +1603,34 @@ void visit_file(char *dir, char *file) /*{{{*/
 
 /*}}}*/
 
-char *dir_file_merge(char *dir, char *file) /*{{{*/
+char *jed_dir_file_merge(SLFUTURE_CONST char *dir, SLFUTURE_CONST char *file) /*{{{*/
 /* 
  * returns result of merging dir with file. If file is empty, dir is
  * considered to be whole file.
  */
 {
-   char name[2 * JED_MAX_PATH_LEN];
+   char *dirfile;
 
-   strcpy (name, dir);
-   if ((file != NULL) && *file)
-     {
-	fixup_dir(name);
-	strcat(name, file);
-     }
+   if ((file == NULL) || (*file == 0))
+     return jed_standardize_filename (dir);
 
-   return expand_filename(name);
+   
+   dirfile = SLmalloc (strlen (dir) + strlen (file) + 2);
+   if (dirfile == NULL)
+     return NULL;
+   strcpy (dirfile, dir);
+   fixup_dir (dirfile);
+   strcat (dirfile, file);
+	
+   file = jed_standardize_filename (dirfile);
+   SLfree (dirfile);
+   return (char *) file;
 }
 
 /*}}}*/
 
-int file_status(char *file) /*{{{*/
+
+int file_status(SLFUTURE_CONST char *file) /*{{{*/
 /*
  *  Returns a coded integer about file.  If the file does not exist, 0 is
  *  returned.  Meaning:
@@ -1280,44 +1643,105 @@ int file_status(char *file) /*{{{*/
  *    -3 unknown error
  */
 {
-   int mode = 0;
-   short uid, gid;
+   mode_t mode = 0;
+   uid_t uid;
+   gid_t gid;
    return sys_chmod(file, 0, &mode, &uid, &gid);
 }
 
 /*}}}*/
-int file_changed_on_disk (Buffer *buf, char *file) /*{{{*/
+
+/* Returns 0 if unchanged, 1 if changed, or -1 if it does not exist
+ * and buffer is unmodified.
+ * 
+ * What does it mean for the file to be changed on the disk if the buffer
+ * corresponds to a symlink?  That is, imagine editing foo.c which is a
+ * symlink to bar.c.  If bar.c changes, then foo.c shold be flagged as
+ * changed.  But, what if the link foo.c is changed such that it nolonger 
+ * points to bar.c?  Now the buffer foo.c is nolonger connected with 
+ * bar.c.
+ * 
+ * To handle both situations, do not use the dirfile field of the buffer
+ * structure, since that corresponds to bar.c and not the link foo.c.
+ * Since the file_changed_on_disk function uses stat and not lstat, if foo.c
+ * is a link to bar.c, then bar.c will be checked.
+ * 
+ * If the file has changed, be sure to update the buffers dirfile field and
+ * update the inode info.  If another buffer has the same inode info, then
+ * the user has shot himself in the foot by making foo.c a hardlink.
+ */
+int file_changed_on_disk (Buffer *buf, char *dirfile) /*{{{*/
 {
    unsigned long t;
 
-   t = sys_file_mod_time(file);
+   /* If we are saving this to a different file, then skip the test --
+    * unless dirfile is a link.
+    */
+   if ((buf->canonical_dirfile != NULL) 
+#if JED_FILE_PRESERVE_CASE
+       && (0 != strcmp (buf->canonical_dirfile, dirfile))
+#else
+       && (0 != jed_case_strcmp (buf->canonical_dirfile, dirfile))
+#endif
+       )
+     {
+#ifdef REAL_UNIX_SYSTEM
+	struct stat s;
+
+	if ((-1 == lstat(dirfile, &s))
+	    || (0 == ((s.st_mode & S_IFMT) & S_IFLNK)))
+	  return 0;
+#else	
+	return 0;
+#endif
+     }
+
+#ifdef REAL_UNIX_SYSTEM
+   /*
+    * The file may still have been changed via a rename operation.
+    * Update (or even invalidate) inode information since directories may have
+    * been renamed, etc...
+    */
+     {
+	int device, inode;
+	(void) jed_get_inode_info (dirfile, &device, &inode);
+	if ((buf->device != -1) && (buf->inode != -1))
+	  {
+	     if ((device != buf->device) || (inode != buf->inode))
+	       buf->flags |= FILE_MODIFIED;
+	  }
+	buf->device = device;
+	buf->inode = inode;
+     }
+#endif
+
+   t = sys_file_mod_time(dirfile);
    if (t == 0)
      {
 	/* File does not exist.  If the buffer has not been modified since
 	 * the last save, then the file has changed on the disk.
 	 */
-#if 1
-	return 0;
-#else
 	if (buf->c_time == 0)
 	  return 0;
 
-	return (0 == (buf->flags & BUFFER_MODIFIED));
-#endif
+	if (buf->flags & BUFFER_MODIFIED)
+	  return 0;
+	return -1;		       /* does not exist */
      }
 
    if (t > buf->c_time)
      {
-	if (buf->c_time == 0)	       /* new buffer, writing to exiting file */
+#if 0  /* Commented out because if the file did not exist when the buffer was created
+	* then c_time will be 0.  But if t>0, then this would indicate that an
+	* external process created/modified the file.
+	*/
+	
+	if (buf->c_time == 0)	       /* new buffer, writing to existing file */
 	  return 0;
-
+#endif
 	return 1;
      }
-   
-   /* FIXME: The file may still have been changed via a rename operation.  Check
-    * the inodes.
-    */
-   
+
    return 0;
 }
 
@@ -1352,19 +1776,16 @@ void auto_save(void) /*{{{*/
 
 void check_buffer(Buffer *b) /*{{{*/
 {
-#ifdef REAL_UNIX_SYSTEM
-   /* Update (or even invalidate) inode information since directories may have
-    * been renamed, etc...
-    */
-   (void) get_inode_info (b->dir, &b->device, &b->inode);
-#endif
-
-   if ((*b->file != 0)
-       && file_changed_on_disk(b, dir_file_merge(b->dir, b->file)))
+   b->flags &= ~FILE_MODIFIED;
+   if (*b->file)
      {
-	b->flags |= FILE_MODIFIED;
+	char *dirfile = jed_dir_file_merge (b->dir, b->file);
+
+	if (file_changed_on_disk (b, dirfile))
+	  b->flags |= FILE_MODIFIED;
+
+	SLfree (dirfile);
      }
-   else b->flags &= ~FILE_MODIFIED;
 }
 
 /*}}}*/
@@ -1390,36 +1811,37 @@ void set_file_trans(int *mode) /*{{{*/
 
 int read_file_pointer(int fp) /*{{{*/
 {
-   int n = 1;
+   int n = 0;
    unsigned int num;
    unsigned char *vbuf;
    VFILE *vp;
+
+   if (SLang_get_error () || (vp = vstream(fp, MAX_LINE_LEN, VFile_Mode)) == NULL) return(-1);
    
-   if (SLang_Error || (vp = vstream(fp, MAX_LINE_LEN, VFile_Mode)) == NULL) return(-1);
-   
-   if (NULL == (vbuf = (unsigned char *) vgets(vp, &num))) return(0);
-   
-#ifdef KEEP_SPACE_INFO
-   if (CLine->space < num)
-#endif
-     remake_line(CLine->len + num + 1);
-   
-   SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
-   CLine->len = num;
-   
-   while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
+   if (NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
      {
 	n++;
-	if ((num == 1) && (*vbuf == '\n')) make_line(num); else make_line(num + 1);
+#ifdef KEEP_SPACE_INFO
+	if (CLine->space < num)
+#endif
+	  remake_line(CLine->len + num + 1);
+	
 	SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
 	CLine->len = num;
-	if (SLang_Error) break;
+   
+	while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
+	  {
+	     n++;
+	     if ((num == 1) && (*vbuf == '\n')) make_line(num); else make_line(num + 1);
+	     SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
+	     CLine->len = num;
+	     if (SLang_get_error () || SLKeyBoard_Quit) break;
+	  }
+	if (vp->cr_flag) CBuf->flags |= ADD_CR_ON_WRITE_FLAG;
+	else CBuf->flags &= ~ADD_CR_ON_WRITE_FLAG;
      }
+   
    if (vp->buf != NULL) SLfree(vp->buf);
-   
-   if (vp->cr_flag) CBuf->flags |= ADD_CR_ON_WRITE_FLAG;
-   else CBuf->flags &= ~ADD_CR_ON_WRITE_FLAG;
-   
    SLfree ((char *)vp);
    return n;
 }
@@ -1427,30 +1849,21 @@ int read_file_pointer(int fp) /*{{{*/
 /*}}}*/
 
 #ifdef REAL_UNIX_SYSTEM
-int get_inode_info (char *dir, int *device, int *inode) /*{{{*/
+
+int jed_get_inode_info (char *dirfile, int *device, int *inode)
 {
    struct stat st;
-   char work [JED_MAX_PATH_LEN + 512];
-   char *f;
-   
-   f = extract_file (dir);
-   if (*f != 0)
-     {
-	unsigned int len = f - dir;
-	strncpy (work, dir, len);
-	work[len] = 0;
-	dir = work;
-     }
    
    *inode = *device = -1;
-   if (-1 == stat (dir, &st)) return -1;
+   if ((dirfile == NULL)
+       || (-1 == stat (dirfile, &st)))
+     return -1;
    
    *inode = st.st_ino;
    *device = st.st_dev;
-   
+
    return 0;
 }
-/*}}}*/
 
 #endif
 
@@ -1472,9 +1885,42 @@ void jed_set_umask (int mask)
 #ifdef REAL_UNIX_SYSTEM
 	int u;
 	
-	u = umask (mask & 0777);
+	u = umask (mask & 07777);
 	if (default_umask == 0) default_umask = u;
 #endif
      }
 }
 
+int jed_dirfile_to_dir_file (char *dirfile, char **dirp, char **filep)
+{
+   char *dir, *file;
+   unsigned int len;
+
+   if (dirfile == NULL)
+     return -1;
+
+   file = extract_file (dirfile);
+   len = (unsigned int) (file - dirfile);
+   if (NULL == (dir = SLmake_nstring (dirfile, len)))
+     return -1;
+   
+   if (filep != NULL)
+     {
+	if (NULL == (file = SLmake_string (file)))
+	  {
+	     SLfree (dir);
+	     return -1;
+	  }
+	*filep = file;
+     }
+
+   *dirp = dir;
+   return 0;
+}
+
+
+char *jed_get_canonical_pathname (char *file)
+{
+   return jed_expand_link (file);
+}
+
