@@ -1,5 +1,5 @@
 /* This is the interface to the GPM mouse under Linux */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006, 2008 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -167,7 +167,7 @@ static int insert_cutbuffer (void)
 {
    CHECK_READ_ONLY
    if (CutBuffer == NULL) return 0;
-   if (CutBuffer_Len) jed_ins_chars ((unsigned char *) CutBuffer, CutBuffer_Len);
+   if (CutBuffer_Len) jed_insert_nbytes ((unsigned char *) CutBuffer, CutBuffer_Len);
    return CutBuffer_Len;
 }
 
@@ -215,7 +215,7 @@ static SLang_Intrin_Fun_Type gpm_mouse_table[] =
    /* Prototype: Void x_copy_region_to_cutbuffer();
     */
    MAKE_INTRINSIC("gpm_disable_mouse", disable_mouse, VOID_TYPE, 0),
-   SLANG_END_TABLE
+   SLANG_END_INTRIN_FUN_TABLE
 };
 
 static void hide_mouse (int show)
@@ -236,9 +236,9 @@ static int open_mouse (void)
     */
    term = getenv ("TERM");
    if ((term != NULL)
-       && !strncmp (term, "xterm", 5))
+       && (!strncmp (term, "xterm", 5) || !strncmp (term, "rxvt", 4)))
      return -1;
-   
+
    /* Another libgpm annoyance.  gpm installs a signal handler for SIGTSTP.
     * I am going to un-install it.
     */
@@ -278,5 +278,3 @@ static int open_mouse (void)
    
 int (*X_Open_Mouse_Hook)(void) = open_mouse;
 void (*X_Close_Mouse_Hook)(void) = close_mouse;
-
-   
