@@ -618,19 +618,17 @@ static int goto_string_end (Syntax_Table_Type *table, unsigned char ch)
  * count is the number of lines to go back
  */
 
-/* FIXME!!! Both the forward and backward got_match routines need to be
- * rewritten to take advantage of the parse_to_point function.
- */
 static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 {
    unsigned char *p, *pmin, want_ch;
    unsigned short *syntax;
-   int in_string, in_comment, level;
+   int in_string, in_comment, in_html, level;
    int quote;
    Syntax_Table_Type *table;
    unsigned int this_syntax;
    unsigned char *pmax;
    unsigned char com_start_char, com_stop_char;
+   unsigned char sgml_start_char, sgml_stop_char;
 
    if (NULL == (table = setup_for_match (&p, &ch, &want_ch)))
      return 0;
@@ -655,6 +653,9 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
       case JED_LINE_IN_STRING1:
 	in_string = (int) table->string_chars[1];
 	break;
+      case JED_LINE_IN_HTML:
+	in_html = 1;
+	break;
      }
    Point--;
 
@@ -667,6 +668,10 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
      com_stop_char = (unsigned char) table->comment_stop[0];
    else com_stop_char = 0;
 
+   sgml_start_char = table->sgml_start_char;
+   sgml_stop_char = table->sgml_stop_char;
+
+   /* FIXME: handle sgml_start/stop_char */
    pmax = CLine->data + CLine->len;
    while (count)
      {
@@ -709,8 +714,15 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 			    /* Match not found.  Leave point at beginning
 			     * of string and return -1.
 			     */
-			    jed_position_point (p + 1);
-			    return -1;
+			    /* Some syntaxes permit double quotes in a string
+			     * to represent a single quote.  Allow that here.
+			     */
+			    if ((p < pmin) || ((int)*p != in_string))
+			      {
+				 jed_position_point (p + 1);
+				 return -1;
+			      }
+			    p--;
 			 }
 		       continue;
 		    }
@@ -954,7 +966,15 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 		  if (in_string)
 		    {
 		       if ((int) ch == in_string)
-			 in_string = 0;
+			 {
+			    /* Some syntaxes permit double quotes in a string
+			     * to represent a single quote.  Allow that here.
+			     */
+			    if ((p >= pmax) || ((int)*p != in_string))
+			      in_string = 0;
+			    else 
+			      p++;     /* skip second quote */
+			 }
 		    }
 		  else
 		    {
@@ -1043,7 +1063,7 @@ int goto_match (void) /*{{{*/
 {
    if (1 != find_matching_delimiter_1 (0, LineNum))
      {
-	if (!IS_MINIBUFFER) msg_error("Mismatch!!");
+	if (!IN_MINI_WINDOW) msg_error("Mismatch!!");
 	return 0;
      }
    return 1;
@@ -1094,7 +1114,7 @@ void blink_match (void) /*{{{*/
 
    if (code == 0)
      {
-	if ((! (CBuf->modes == WRAP_MODE)) && (!IS_MINIBUFFER))
+	if ((! (CBuf->modes == WRAP_MODE)) && (!IN_MINI_WINDOW))
 	  message("Mismatch??");
      }
    else if ((code == 1) && is_line_visible (LineNum))
