@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -20,8 +20,13 @@
 #define JED_HAS_BUFFER_LOCAL_VARS	1
 #define JED_HAS_SAVE_NARROW		1
 
+#define JED_HAS_DISPLAY_LINE_NUMBERS	1
+
 /* Drop down menu support. */
-#define JED_HAS_TTY_MENUS		1
+#define JED_HAS_MENUS		1
+
+/* See ../doc/filelock.txt --- requires symbolic links */
+#define JED_HAS_EMACS_LOCKING		1
 
 /* Double/Triple click support.  This is currently supported by:
  *
@@ -50,15 +55,38 @@
 
 /* Enhanced syntax highlighting support.  This is a much more sophisticated
  * approach based on regular expressions.  Experimental.
- * 
- * Right now, DFA syntax highlighting is broken.  Do NOT change this unless
- * you first fix it!!!  Be sure to send me the patches.
  */
-#define JED_HAS_DFA_SYNTAX		0
+#define JED_HAS_DFA_SYNTAX		1
 
 /* Set JED_HAS_ABBREVS to 1 for the abbreviation feature. */
 #define JED_HAS_ABBREVS			1
 #define JED_HAS_COLOR_COLUMNS		1
 #define JED_HAS_LINE_MARKS		1
 
+/* The following only applies to systems that have case-insensitive file
+ * systems.
+ *
+ * Are filenames case sensitive? If this is set to 1 Ma* will not match
+ * makefile, even on systems that are not case sensitive by default. 
+ */
+#define	JED_FILENAME_CASE_SENSITIVE	0
+
+/* If compiled with XJED_SET_WM_COMMAND non-zero, then Xjed
+ * will set the WM_COMMAND property to inform the
+ * window manager or session manager of the command line used to
+ * start the application.  Not all versions of X support this.
+ */
+#define XJED_SET_WM_COMMAND	0
+
+#if JED_HAS_EMACS_LOCKING && !defined(HAVE_SYMLINK)
+# undef JED_HAS_EMACS_LOCKING
+# define JED_HAS_EMACS_LOCKING 0
+#endif
+
+/* Set JED_HAS_IMPORT if you want the ability to import modules into jed via
+ * the slang import statement.  This assumes that slang was compiled with 
+ * such support.
+ */
+#define JED_HAS_IMPORT	1
+
 #endif
