@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -10,6 +10,10 @@
 
 #include <stdio.h>
 
+#if defined(__QNX__) && defined(__WATCOMC__)
+# include <env.h>
+#endif
+
 #if JED_HAS_SUBPROCESSES
 /* Everything else here is in this '#if' */
 
@@ -53,7 +57,7 @@
 #endif
 
 #ifdef REALLY_HAVE_TERMIOS_H
-# ifndef __os2__
+# if !defined(__os2__) && !defined(__APPLE__)
 #  define USE_PTY
 # endif
 #endif
@@ -64,11 +68,14 @@
 #include "misc.h"
 #include "jprocess.h"
 #include "paste.h"
+#include "sysdep.h"
 #include "sig.h"
 #include "cmds.h"
 
 /*}}}*/
 
+#define MAX_TTY_SLAVE_NAME	256
+
 int Num_Subprocesses;
 int Max_Subprocess_FD;
 
@@ -76,7 +83,8 @@ int Max_Subprocess_FD;
  * pseudo-pids used here.  See, e.g., jed_get_child_status for usage in this
  * fashion.
  */
-int Subprocess_Read_fds [MAX_PROCESSES][2];   /* 0 is actual fd, 1 is our rep */
+int Subprocess_Read_fds [MAX_PROCESSES][3];   
+/* 0 is actual fd, 1 is our rep, 2 will be set to 1 if an EIO is present */
 
 volatile int Child_Status_Changed_Flag;/* if this is non-zero, editor
 					* should call the appropriate
@@ -106,14 +114,20 @@ typedef struct /*{{{*/
    Buffer *buffer;		       /* buffer associated with process */
    SLang_Name_Type *slang_fun;	       /* function to pass output to */
    SLang_MMT_Type *umark;	       /* marks point of last output */
-   
+   int process_flags;
+#define USE_CURRENT_BUFFER	0x1    /* use the current buffer instead of
+					* the one associated with the process
+					*/
+
    SLang_Name_Type *status_change_fun; /* call this if process status changes 
 					* The function should be declared like
-					* define fun (pid, flags);
+					* define fun (pid, flags, status);
 					* The flags parameter corresponds to
 					* the flags field in this struct and 
-					* the pid is NOT the pid of this struct
+					* the pid is NOT the pid of this struct.
+					* status depends upon flags.
 					*/
+   int quietly_kill_on_exit;
 } 
 /*}}}*/
 Process_Type;
@@ -199,7 +213,7 @@ static Process_Type *get_process (int fd) /*{{{*/
        && (p = &Processes[fd], p->flags != 0))
      return p;
    
-   msg_error ("process does not exist.");
+   jed_verror ("process '%d' does not exist.", fd);
    return NULL;
 }
 
@@ -211,12 +225,15 @@ static void call_slang_status_change_hook (Process_Type *p) /*{{{*/
    if ((p->status_change_fun == NULL) || (p->buffer == NULL)) return;
    
    cbuf->locked++;
-   switch_to_buffer (p->buffer);
+   if (0 == (p->process_flags & USE_CURRENT_BUFFER))
+     switch_to_buffer (p->buffer);
    SLang_push_integer ((int) (p - Processes));
    SLang_push_integer (p->flags);
+   SLang_push_integer (p->return_status);
    SLexecute_function (p->status_change_fun);
    touch_screen ();
-   if (CBuf != cbuf) switch_to_buffer (cbuf);
+   if ((0 == (p->process_flags & USE_CURRENT_BUFFER)) && (CBuf != cbuf))
+     switch_to_buffer (cbuf);
    cbuf->locked--;
 }
 
@@ -232,6 +249,23 @@ int jed_signal_process (int *fd, int *sig) /*{{{*/
    return 0;
 }
 
+int jed_signal_fg_process (int *fd, int *sig) /*{{{*/
+{
+   int pid;
+
+   Process_Type *p;
+   if (NULL == (p = get_process (*fd))) return -1;
+
+#ifdef HAVE_TCGETPGRP
+   pid = tcgetpgrp (p->rd);
+   if (pid == -1)
+#endif
+     pid = p->pid;
+
+   (void) kill (-pid, *sig);
+   return 0;
+}
+
 /*}}}*/
 #endif
 
@@ -274,15 +308,22 @@ static void get_process_status (Process_Type *p) /*{{{*/
    slfd = (int) (p - Processes);
 
    if (p->umark != NULL) SLang_free_mmt (p->umark);
+
+#if SLANG_VERSION > 10400
+   SLang_free_function (p->slang_fun);
+   SLang_free_function (p->status_change_fun);
+#endif
+
    memset ((char *)p, 0, sizeof (Process_Type));
-   
+   p->rd = p->wd = -1;
+
    /* Adjust the array of read descriptors */
 
    i = 0;
    
    while (i < Num_Subprocesses)
      {
-	if (Subprocess_Read_fds[i][0] == slfd)
+	if (Subprocess_Read_fds[i][1] == slfd)
 	  break;
 	  
 	i++;
@@ -294,6 +335,7 @@ static void get_process_status (Process_Type *p) /*{{{*/
      {
 	Subprocess_Read_fds[i][0] = Subprocess_Read_fds[i + 1][0];
 	Subprocess_Read_fds[i][1] = Subprocess_Read_fds[i + 1][1];
+	Subprocess_Read_fds[i][2] = Subprocess_Read_fds[i + 1][2];
 	i++;
      }
    
@@ -353,6 +395,16 @@ void jed_get_child_status (void) /*{{{*/
 {
    Process_Type *p, *pmax;
    
+   Ignore_User_Abort++;
+   if (SLang_get_error ())
+     {
+	Ignore_User_Abort--;
+	return;
+     }
+
+   /* FIXME: We should block SIGCHLD before trying to access this flag,
+    * or anything in this function for that matter.
+    */
    Child_Status_Changed_Flag--;
    
    get_process_input (&Number_Zero);
@@ -369,6 +421,9 @@ void jed_get_child_status (void) /*{{{*/
 	  }
 	p++;
      }
+
+   Ignore_User_Abort--;
+
 }
 
 /*}}}*/
@@ -376,56 +431,65 @@ void jed_get_child_status (void) /*{{{*/
 static void child_signal_handler (int sig) /*{{{*/
 {
    int status;
-   int return_status;
-   int pid;
-   Process_Type *p, *pmax;
+   int found;
    int save_errno = errno;
    
    (void) sig;
 
-   while (1)
+   do
      {
-	pid = (int) waitpid (-1, &status, WNOHANG | WUNTRACED);
-	if (pid == -1) 
-	  {
-	     if (errno == ECHILD) break;
-	     continue;
-	  }
-	
-	if (pid == 0) break;
-	
-	return_status = 0;
-	
-	if (WIFEXITED (status))
-	  {
-	     return_status = WEXITSTATUS (status);
-	     status = PROCESS_EXITED;
-	  }
-	else if (WIFSIGNALED (status))
-	  {
-	     status = PROCESS_SIGNALLED;
-	     return_status = WTERMSIG (status);
-	  }
-	else if (WIFSTOPPED (status))
-	  {
-	     status = PROCESS_STOPPED;
-	  }
-	/* What else?? */
-	
+	Process_Type *p, *pmax;
+
+	found = 0;
+
 	p = Processes;
 	pmax = p + MAX_PROCESSES;
 	while (p < pmax)
 	  {
+	     int pid;
+	     
+	     if ((pid = p->pid) <= 0)
+	       {
+		  p++;
+		  continue;
+	       }
+
+	     while ((-1 == (pid = (int) waitpid (p->pid, &status, WNOHANG | WUNTRACED)))
+		    && (errno == EINTR))
+	       ;
+
 	     if (p->pid == pid)
 	       {
+		  int return_status;
+
+		  found++;
+
+		  if (WIFEXITED (status))
+		    {
+		       return_status = WEXITSTATUS (status);
+		       status = PROCESS_EXITED;
+		    }
+		  else if (WIFSIGNALED (status))
+		    {
+		       return_status = WTERMSIG (status);
+		       status = PROCESS_SIGNALLED;
+		    }
+#ifdef WIFSTOPPED
+		  else if (WIFSTOPPED (status))
+		    {
+		       return_status = WSTOPSIG(status);
+		       status = PROCESS_STOPPED;
+		    }
+#endif
 		  p->flags = status;
 		  p->status_changed++;
 		  p->return_status = return_status;
-		  break;
 	       }
 	     p++;
 	  }
      }
+   while (found != 0);
+
    SLsignal_intr (SIGCHLD, child_signal_handler);
    errno = save_errno;
    Child_Status_Changed_Flag++;
@@ -439,40 +503,46 @@ static void child_signal_handler (int sig) /*{{{*/
 
 static int get_master_slave_fds (int *slave_read, int *slave_write,
 				 int *master_read, int *master_write,
-				 char *slave_tty_name, int *is_pty)
+				 char *slave_tty_name, int *is_pty,
+				 int want_pty)
 {
 #ifdef USE_PTY
-   int master;
-   
-   if (-1 == pty_open_master_pty (&master, slave_tty_name))
-     return -1;
-
-   *master_read = *master_write = master;
-   *slave_read = *slave_write = -1;
-   
-   *is_pty = 1;
-   return 0;
-#else
-
-   int fds0[2], fds1[2];
-   
-   if (-1 == pipe (fds0))
-     return -1;
-   
-   if (-1 == pipe (fds1))
+   if (want_pty)
      {
-	signal_safe_close (fds0[0]);
-	signal_safe_close (fds0[1]);
-	return -1;
+	int master;
+	
+	if (-1 == pty_open_master_pty (&master, slave_tty_name))
+	  return -1;
+
+	*master_read = *master_write = master;
+	*slave_read = *slave_write = -1;
+	
+	*is_pty = 1;
      }
-   *master_read = fds0[0];
-   *slave_write = fds0[1];
-   *master_write = fds1[1];
-   *slave_read = fds1[0];
-   *slave_tty_name = 0;
-   *is_pty = 0;
-   return 0;
+   else
+     {
 #endif
+	int fds0[2], fds1[2];
+   
+	if (-1 == pipe (fds0))
+	  return -1;
+	
+	if (-1 == pipe (fds1))
+	  {
+	     signal_safe_close (fds0[0]);
+	     signal_safe_close (fds0[1]);
+	     return -1;
+	  }
+	*master_read = fds0[0];
+	*slave_write = fds0[1];
+	*master_write = fds1[1];
+	*slave_read = fds1[0];
+	*slave_tty_name = 0;
+	*is_pty = 0;
+#ifdef USE_PTY
+     }
+#endif
+   return 0;
 }
 
 
@@ -483,7 +553,7 @@ static void my_setenv (char *what, char *value)
 #else
 # ifdef HAVE_PUTENV
    char buf[512];
-   sprintf (buf, "%s=%s", what, value);
+   SLsnprintf (buf, sizeof (buf), "%s=%s", what, value);
    (void) putenv (buf);
 # endif
 #endif
@@ -498,7 +568,7 @@ static void my_unsetenv (char *what)
 
    
 #ifdef __os2__
-static int open_process (char *pgm, char **argv) /*{{{*/
+static int open_process (char *pgm, char **argv, int want_pty) /*{{{*/
 {
     int val;
     int pd;
@@ -506,7 +576,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    int pid, i;
    Process_Type *p;
    SLang_MMT_Type *mmt;
-   char slave_tty_name [80];
+   char slave_tty_name [MAX_TTY_SLAVE_NAME];
    int max_write_tries;
    SLSig_Fun_Type *sig_orig[NSIG];
 
@@ -527,7 +597,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    
    if (-1 == get_master_slave_fds (&slave_read, &slave_write,
 				   &master_read, &master_write,
-				   slave_tty_name, &p->is_pty))
+				   slave_tty_name, &p->is_pty, want_pty))
      {
 	SLang_free_mmt (mmt);
 	return -1;
@@ -602,6 +672,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    
    Subprocess_Read_fds[Num_Subprocesses][0] = master_read;
    Subprocess_Read_fds[Num_Subprocesses][1] = pd;
+   Subprocess_Read_fds[Num_Subprocesses][2] = 0;
    if (master_read > Max_Subprocess_FD) Max_Subprocess_FD = master_read;
    Num_Subprocesses += 1;
    
@@ -665,8 +736,12 @@ static void wait_for_parent (void)
     */
    while (SigUsr_Flag == 0)
      sigsuspend (&Zero_Signal_Mask);
-   
+}
+
+static void tell_parent_to_go (pid_t pid)
+{
    SigUsr_Flag = 0;
+   kill (pid, SIGUSR1);
    
    while ((-1 == sigprocmask (SIG_SETMASK, &Old_Signal_Mask, NULL))
 	  && (errno == EINTR))
@@ -675,9 +750,13 @@ static void wait_for_parent (void)
 
 static void tell_child_to_go (pid_t pid)
 {   
+   SigUsr_Flag = 0;
    kill (pid, SIGUSR1);
 
-   SigUsr_Flag = 0;
+   /* Now wait for the child to setup the pty */
+   while (SigUsr_Flag == 0)
+     sigsuspend (&Zero_Signal_Mask);   
+
    while ((-1 == sigprocmask (SIG_SETMASK, &Old_Signal_Mask, NULL))
 	  && (errno == EINTR))
      ;
@@ -691,14 +770,21 @@ static int set_non_blocking_io (int fd)
    return signal_safe_fcntl (fd, F_SETFL, val | O_NONBLOCK);
 }
 
-static int open_process (char *pgm, char **argv) /*{{{*/
+#ifdef SIGTRAP
+static void sigtrap_handler (int sig)
+{
+   (void) sig;
+}
+#endif
+   
+static int open_process (char *pgm, char **argv, int want_pty) /*{{{*/
 {
    int pd;
    int slave_read, slave_write, master_read, master_write;
-   int pid, i;
+   int pid, jed_pid, i;
    Process_Type *p;
    SLang_MMT_Type *mmt;
-   char slave_tty_name [80];
+   char slave_tty_name [MAX_TTY_SLAVE_NAME];
 
    if (Jed_Secure_Mode)
      {
@@ -717,7 +803,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    
    if (-1 == get_master_slave_fds (&slave_read, &slave_write,
 				   &master_read, &master_write,
-				   slave_tty_name, &p->is_pty))
+				   slave_tty_name, &p->is_pty, want_pty))
      {
 	SLang_free_mmt (mmt);
 	return -1;
@@ -725,6 +811,8 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 
    SLsignal_intr (SIGCHLD, child_signal_handler);
 
+   jed_pid = getpid ();		       /* used by slave */
+
    if ((-1 == init_child_parent_sync ())
        || ((pid = fork ()) < 0))
      {
@@ -759,8 +847,6 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 	  signal_safe_close (master_write);	       /* close write end of 0 */
 	signal_safe_close (master_read);	       /* close read end of 1 */
 
-	for (i = 0; i < 32; i++) SLsignal (i, SIG_DFL);
-	
 #ifdef USE_PTY	
 	/* Call set setsid so that the child will become the session leader.
 	 * This has the side effect that we will loose the controlling
@@ -775,6 +861,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 	
 	     if (-1 == pty_open_slave_pty (slave_tty_name, &slave_read))
 	       {
+		  tell_parent_to_go (jed_pid);
 		  fprintf (stderr, "child: failed to open slave.");
 		  _exit (1);
 	       }
@@ -783,6 +870,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 	     /* (void) pty_setup_slave_term (slave_read, 0); */
 	  }
 #endif	/* USE_PTY */
+	tell_parent_to_go (jed_pid);
 
 #ifdef USE_PTY
 	/* Put tty back into cbreak mode. */
@@ -799,6 +887,21 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 	my_setenv ("TERM", "unknown");
 	my_unsetenv ("TERMCAP");
 
+	for (i = 0; i < 32; i++) 
+	  {
+#ifdef SIGTRAP
+	/* Under Linux, subprocesses fail if run under the debugger because
+	 * something sends this child a SIGTRAP.  Let's try ignoring it.
+	 */
+	     if (i == SIGTRAP)
+	       SLsignal (i, sigtrap_handler);
+	     else
+#endif
+	       SLsignal (i, SIG_DFL);
+	  }
+#ifdef SIGTRAP
+	SLsignal (SIGTRAP, SIG_DFL);
+#endif
 	if (execvp (pgm, argv) < 0)
 	  {
 	     fprintf (stderr, "execvp of %s failed!\r\n", pgm);
@@ -819,6 +922,7 @@ static int open_process (char *pgm, char **argv) /*{{{*/
    
    Subprocess_Read_fds[Num_Subprocesses][0] = master_read;
    Subprocess_Read_fds[Num_Subprocesses][1] = pd;
+   Subprocess_Read_fds[Num_Subprocesses][2] = 0;
    if (master_read > Max_Subprocess_FD) Max_Subprocess_FD = master_read;
    Num_Subprocesses += 1;
    
@@ -841,10 +945,32 @@ static int open_process (char *pgm, char **argv) /*{{{*/
 /*}}}*/
 #endif				       /* __os2__ */
 
+static int flag_fd_as_eio_error (int fd)
+{
+   unsigned int i, n;
+   
+   n = Num_Subprocesses;
+   for (i = 0; i < n; i++)
+     {
+	if (Subprocess_Read_fds[i][0] == fd)
+	  {
+	     Subprocess_Read_fds[i][2] = 1;
+	     return 0;
+	  }
+     }
+   return -1;
+}
 
 /* This function is only called when we are reading characters from the 
  * keyboard.  Keyboard input has the highest priority and this is called only
  * if there is no input ready.
+ * 
+ * It might also get called by the hooks that it calls.  As a result, it must
+ * be reentrant.
+ * 
+ * It is possible for the user to press Ctrl-G while this routine is executing.
+ * This could cause slang hooks to fail.  It would probably be better to 
+ * block SIGINT for the duration of this routine.
  */
 void read_process_input (int fd) /*{{{*/
 {
@@ -860,15 +986,23 @@ void read_process_input (int fd) /*{{{*/
 	return;
      }
    
+   Ignore_User_Abort++;
+   if (SLang_get_error ())
+     {
+	Ignore_User_Abort--;
+	return;
+     }
+   
    otype = p->output_type;
    pbuf = p->buffer;
    
    if (pbuf != NULL)
      {
-	switch_to_buffer (pbuf);
+	if (0 == (p->process_flags & USE_CURRENT_BUFFER))
+	  switch_to_buffer (pbuf);
 	pbuf->locked++;
      }
-   
+
    total = 0;
    if (otype & PROCESS_SAVE_POINT) push_spot ();
    while ((n = read (p->rd, buf, 512)) > 0)
@@ -879,26 +1013,45 @@ void read_process_input (int fd) /*{{{*/
 	if (otype & PROCESS_USE_BUFFER) 
 	  {
 	     if (0 == (otype & PROCESS_AT_POINT)) eob ();
-	     ins_chars (buf, n);
+	     (void) jed_insert_nbytes (buf, n);
 	     jed_move_user_object_mark (p->umark);
 	  }
 	else if (otype == PROCESS_USE_SLANG)
 	  {
 	     buf[n] = 0;
-	     SLang_push_integer ((int) (p - Processes));
-	     SLang_push_string ((char *) buf);
-	     SLexecute_function (p->slang_fun);    /* function to pass output to */
+	     if ((-1 == SLang_push_integer ((int) (p - Processes)))
+		 || (-1 == SLang_push_string ((char *) buf))
+		 || (-1 == SLexecute_function (p->slang_fun)))
+	       {
+		  /* It looks like the contents of buf may be lost */
+		  break;
+	       }
 	  }
      }
    
+   if (n == -1)
+     {
+#ifdef EIO
+	if (errno == EIO)
+	  (void) flag_fd_as_eio_error (p->rd);
+#endif
+     }
+
    if (otype & PROCESS_SAVE_POINT) pop_spot ();
    else if (otype & PROCESS_USE_BUFFER) move_window_marks (0);
    
    if (p->buffer != NULL)
      {
-	if (b != CBuf) switch_to_buffer (b);
-	pbuf->locked--;
+	if ((b != CBuf) && (0 == (p->process_flags & USE_CURRENT_BUFFER)))
+	  switch_to_buffer (b);
      }
+   
+   /* Since it was locked, it cannot be deleted.  So pbuf is still a good 
+    * pointer
+    */
+   if (pbuf != NULL)
+     pbuf->locked--;
+
    if (total)
      {
 #ifdef __os2__
@@ -908,6 +1061,8 @@ void read_process_input (int fd) /*{{{*/
 #endif
 	touch_screen ();
      }
+   
+   Ignore_User_Abort--;
 }
 
 
@@ -915,17 +1070,47 @@ void read_process_input (int fd) /*{{{*/
 
 static int write_to_process (int fd, char *buf, unsigned int len)
 {
+   int n, ilen;
+   
    if (len == 0) 
      return 0;
    
-   /* FIXME!! check for write errors!! */
-   while (-1 == write (fd, buf, len))
+   ilen = (int) len;
+   n = 0;
+   while (n < ilen)
      {
-	if (errno != EINTR)
+	int dn;
+	int count = 0;
+
+	while (-1 == (dn =  write (fd, buf + n, ilen - n)))
+	  {
+	     if (errno == EINTR)
+	       continue;
+	     
+#ifdef EAGAIN
+	     if (errno == EAGAIN)
+	       {
+		  get_process_input (&Number_Zero);
+		  count++;
+		  if ((count % 10) == 0)
+		    {
+		       sleep (1);
+		       if (count > 20)
+			 return n;
+		    }
+		  continue;
+	       }
+#endif
+	     return -1;
+	  }
+
+	if (dn == 0)
 	  break;
+
+	n += dn;
      }
-   
-   return 0;
+  
+   return n;
 }
 
 int jed_send_process (int *fd, char *str) /*{{{*/
@@ -936,7 +1121,10 @@ int jed_send_process (int *fd, char *str) /*{{{*/
    
    len = strlen (str);
    
-   return write_to_process (p->wd, str, len);
+   if (len != (unsigned int) write_to_process (p->wd, str, len))
+     jed_verror ("write to process failed");
+   
+   return len;
 }
 
 /*}}}*/
@@ -963,34 +1151,106 @@ void jed_send_process_eof (int *fd) /*{{{*/
 
 /*}}}*/
 
-void jed_set_process (int *pd, char *what, char *s) /*{{{*/
+/* If s is NULL, then f != NULL, or visa-versa */
+static int set_process (Process_Type *p, char *what, char *s, SLang_Name_Type *f) /*{{{*/
 {
-   Process_Type *p;
-   SLang_Name_Type *f;
-   
-   if (NULL == (p = get_process (*pd))) return;
    if (!strcmp (what, "output"))
      {
-	if (*s == '.') p->output_type = PROCESS_AT_POINT | PROCESS_USE_BUFFER;
-	else if (*s == '@') p->output_type = PROCESS_SAVE_POINT | PROCESS_USE_BUFFER;
-	else if (*s && (NULL != (f = SLang_get_function (s))))
+	if (s != NULL)
 	  {
-	     p->output_type = PROCESS_USE_SLANG;
-	     p->slang_fun = f;
+	     if (0 == strcmp (s, "."))
+	       {
+		  p->output_type = PROCESS_AT_POINT | PROCESS_USE_BUFFER;
+		  return 0;
+	       }
+	     
+	     if (0 == strcmp (s, "@"))
+	       {
+		  p->output_type = PROCESS_SAVE_POINT | PROCESS_USE_BUFFER;
+		  return 0;
+	       }
+	     
+	     if (*s == 0)
+	       {
+		  p->output_type = PROCESS_USE_BUFFER;
+		  return 0;
+	       }
+
+	     if (NULL == (f = SLang_get_function (s)))
+	       return -1;
 	  }
-	else p->output_type = PROCESS_USE_BUFFER;
+	
+	p->output_type = PROCESS_USE_SLANG;
+	p->slang_fun = f;
+	return 0;
      }
    else if (!strcmp (what, "signal"))
      {
-	if (*s && (NULL != (f = SLang_get_function (s))))
-	  {
-	     p->status_change_fun = f;
-	  }
+	if ((s != NULL)
+	    && (NULL == (f = SLang_get_function (s))))
+	  return -1;
+
+	p->status_change_fun = f;
+	return 0;
      }
+   
+   jed_verror ("set_process: %s not supported", what);
+   return -1;
 }
 
 /*}}}*/
 
+void jed_set_process (void)
+{
+   int pd;
+   char *what, *s;
+   Process_Type *p;
+   SLang_Name_Type *f;
+
+   f = NULL;
+   what = NULL;
+   s = NULL;
+
+   if (SLang_peek_at_stack () == SLANG_REF_TYPE)
+     {
+	if (NULL == (f = SLang_pop_function ()))
+	  return;
+     }
+   else if (-1 == SLang_pop_slstring (&s))
+     return;
+   
+   if ((0 == SLang_pop_slstring (&what))
+       && (0 == SLang_pop_integer (&pd))
+       && (NULL != (p = get_process (pd)))
+       && (0 == set_process (p, what, s, f)))
+     f = NULL;
+     
+   SLang_free_slstring (what);
+   SLang_free_slstring (s);
+#if SLANG_VERSION > 10400
+   SLang_free_function (f);
+#endif
+}
+
+
+void jed_set_process_flags (int *fd, int *oflags)
+{
+   Process_Type *p;
+   if (NULL == (p = get_process (*fd))) 
+     return;
+   
+   p->process_flags = *oflags;
+}
+ 
+int jed_get_process_flags (int *fd)
+{
+   Process_Type *p;
+   if (NULL == (p = get_process (*fd))) 
+     return -1;
+
+   return p->process_flags;
+}
+
 void jed_get_process_mark (int *fd) /*{{{*/
 {
    Process_Type *p;
@@ -1001,11 +1261,11 @@ void jed_get_process_mark (int *fd) /*{{{*/
 
 /*}}}*/
 
-int jed_open_process (int *np) /*{{{*/
+static int open_process_of_type (int nargs, int wantpty) /*{{{*/
 {
    int fd = -1;
    char *argv[502];
-   int n = *np;
+   int n = nargs;
 
    if (CBuf->subprocess)
      {
@@ -1032,16 +1292,16 @@ int jed_open_process (int *np) /*{{{*/
 	  }
      }
    n = 0;
-   
-   if ((fd = open_process(argv[0], argv)) < 0)
+
+   if ((fd = open_process(argv[0], argv, wantpty)) < 0)
      {
-	msg_error ("Unable to open process.");
+	jed_verror ("Unable to open %s process.", argv[0]);
      }
    
    /* free up the argument strings */
    free_return:
    
-   while (n <= *np)
+   while (n <= nargs)
      {
 	SLang_free_slstring (argv[n]);
 	n++;
@@ -1052,6 +1312,17 @@ int jed_open_process (int *np) /*{{{*/
 
 /*}}}*/
 
+int jed_open_process (int *np)
+{
+   return open_process_of_type (*np, 1);
+}
+
+int jed_open_process_pipe (int *np)
+{
+   return open_process_of_type (*np, 0);
+}
+
+
 void jed_block_child_signal (int block) /*{{{*/
 {
    static sigset_t new_mask, old_mask;
@@ -1188,8 +1459,9 @@ int jed_pclose(FILE *fp) /*{{{*/
    int ret;
 
    fd = fileno(fp);
-   if ((fd >= OPEN_MAX) ||
-       (0 == (pid = Popen_Child_Pids[fd]))) 
+   if ((fd >= OPEN_MAX)
+       || (fd < 0)
+       || (0 == (pid = Popen_Child_Pids[fd])))
      return -1;
 
    Popen_Child_Pids [fd] = 0;
@@ -1197,10 +1469,12 @@ int jed_pclose(FILE *fp) /*{{{*/
    if (fclose(fp) == EOF) return(-1);
 
    /* This is the part that the SunOS pclose was apparantly screwing up. */
-   while (pid != (ret = waitpid(pid, &stat, 0)))
+   while (-1 == waitpid(pid, &stat, 0))
      {
-	if ((ret == -1) && (errno != EINTR))
+	if (errno != EINTR)
 	  return -1;
+
+	(void) SLang_handle_interrupt ();
      }
    
    ret = WEXITSTATUS(stat);
@@ -1213,4 +1487,88 @@ int jed_pclose(FILE *fp) /*{{{*/
 /*}}}*/
 #endif
 
+
+void jed_query_process_at_exit (int *pid, int *query)
+{
+   Process_Type *p;
+
+   if (NULL == (p = get_process (*pid)))
+     return;
+   
+   p->quietly_kill_on_exit = (*query == 0);
+}
+
+   
+int jed_processes_ok_to_exit (void)
+{     
+   Process_Type *p, *pmax;
+   int num;
+   char buf[64];
+
+   if (Num_Subprocesses == 0)
+     return 1;
+
+   num = 0;
+   p = Processes;
+   pmax = p + MAX_PROCESSES;
+   
+   while (p < pmax)
+     {
+	if ((p->flags & PROCESS_ALIVE)
+	    && (0 == p->quietly_kill_on_exit))
+	  num++;
+	p++;
+     }
+   
+   if (num == 0)
+     return 1;
+   
+   sprintf (buf, "%d Subprocesses exist.  Exit anyway", num);
+   return jed_get_y_n (buf);
+}
+
+int jed_fork_monitor (void)
+{
+   pid_t pid;
+   int i;
+
+   while (-1 == (pid = fork ()))
+     {
+	if (errno == EINTR)
+	  continue;
+
+	fprintf (stderr, "Unable to fork: errno=%d\n", errno);
+	return -1;
+     }
+   
+   if (pid != 0)
+     return 0;
+   
+   pid = getppid ();
+
+   for (i = 1; i < 32; i++)
+     (void) SLsignal (i, SIG_IGN);
+
+   SLsignal (SIGQUIT, SIG_DFL);
+   SLsignal (SIGTERM, SIG_DFL);
+   SLsignal (SIGCONT, SIG_DFL);
+   SLsignal (SIGTSTP, SIG_DFL);
+   
+   for (i = 0; i < 256; i++)
+     {
+	while ((-1 == close (i)) 
+	       && (errno == EINTR))
+	  ;
+     }
+
+   while (1)
+     {
+	if ((-1 == kill (pid, 0))
+	    && (errno == ESRCH))
+	  _exit (0);
+	
+	sleep (10);
+     }
+}
+
 #endif 				       /* JED_HAS_SUBPROCESSES */
