@@ -16,7 +16,7 @@
 /* extern CRITICAL_SECTION Critical_Section; */
 extern HANDLE Input_Events[];
 #else
-extern int Subprocess_Read_fds [MAX_PROCESSES][2];
+extern int Subprocess_Read_fds [MAX_PROCESSES][3];
 extern int Max_Subprocess_FD;
 #endif
 
@@ -40,9 +40,14 @@ extern int jed_signal_fg_process (int *, int *);
 extern int jed_processes_ok_to_exit (void);
 extern void jed_query_process_at_exit (int *, int *);
 
+extern void jed_set_process_flags (int *, int *);
+extern int jed_get_process_flags (int *);
+
 #ifdef REAL_UNIX_SYSTEM
 extern void jed_block_child_signal (int);
+extern int jed_fork_monitor (void);
 #endif
 
 extern FILE *jed_popen (char *, char *);
 extern int jed_pclose (FILE *);
+
