@@ -1,5 +1,5 @@
-/* Drop-down menus */
-/* Copyright (c) 1999, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
+/* Drop-down menus (tty) */
+/* Copyright (c) 1999, 2000, 2002-2009 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -9,7 +9,7 @@
 #include "jed-feat.h"
 
 #include <stdio.h>
-#if JED_HAS_TTY_MENUS
+#if JED_HAS_MENUS
 
 #include <slang.h>
 
@@ -968,41 +968,41 @@ static void draw_down_arrow (int r, int c)
 
 static void draw_name (char *name, int color0, int color1, unsigned int field_width)
 {
-   char *s;
-   unsigned int n;
+   unsigned char *s, *smax;
+   unsigned int name_width;
 
-   s = name;
+   s = (unsigned char *) name;
+   smax = s + strlen (name);
+#if SLANG_VERSION >= 20000
+   name_width = SLsmg_strwidth (s, smax);
+#else
+   name_width = smax - s;
+#endif
 
-   while (*s && (*s != '&'))
+   while ((s < smax) && (*s != '&'))
      s++;
 
-   n = (unsigned int) (s - name);
-   if (n)
-     {
-	SLsmg_set_color (color0);
-	SLsmg_write_nchars (name, n);
-     }
+   SLsmg_set_color (color0);
+   SLsmg_write_chars ((unsigned char *) name, s);
 
-   if (*s != 0)
+   if (s < smax)
      {
-	unsigned int dn;
-
-	s++;
+	unsigned char *s1;
+	
+	name_width--;
+	s++;			       /* skip & */
 	SLsmg_set_color (color1);
-	SLsmg_write_nchars (s, 1);
-	n++;
+	s1 = jed_multibyte_chars_forward (s, smax, 1, NULL, 1);
+	SLsmg_write_chars (s, s1);
 	SLsmg_set_color (color0);
-	if (*s != 0)
-	  {
-	     s++;
-	     dn = strlen (s);
-	     SLsmg_write_nchars (s, dn);
-	     n += dn;
-	  }
+	s = s1;
+
+	if (s < smax)
+	  SLsmg_write_chars (s, smax);
      }
 
-   if (n < field_width)
-     SLsmg_write_nstring ("", field_width - n);
+   if (name_width < field_width)
+     SLsmg_write_nstring ("", field_width - name_width);
 }
 
 	  
@@ -1330,7 +1330,7 @@ static Menu_Bar_Type *get_active_menubar (void)
      {
 	Buffer *buf = CBuf;
 
-	if (IS_MINIBUFFER)
+	if (IN_MINI_WINDOW)
 	  {
 	     if (NULL == (buf = jed_get_mini_action_buffer ()))
 	       return NULL;
@@ -1902,7 +1902,7 @@ int jed_select_menu_bar (void)
    Menu_Popup_Type *p;
    Menu_Node_Type **l, **lmax;
 
-   if (IS_MINIBUFFER)
+   if (IN_MINI_WINDOW)
      {
 	/* For now do not allow access to menus from the minibuffer
 	 * since the minibuffer is not recursive
@@ -2604,4 +2604,4 @@ static int xterm_mouse_cmd (void)
 
 #endif 				       /* !IBMPC_SYSTEM */
 #endif				       /* HAS_MOUSE */
-#endif				       /* JED_HAS_TTY_MENUS */
+#endif				       /* JED_HAS_MENUS */
