@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -13,6 +13,8 @@
 #include <string.h>
 #include <slang.h>
 
+#ifdef HAS_MOUSE
+
 #include "jdmacros.h"
 
 #include "buffer.h"
@@ -62,11 +64,14 @@ static JMouse_Type Current_Event;
 static int Mouse_Queue_Hint;
 
 
-static unsigned char Mouse_Button_Map [3] =
+static unsigned char Mouse_Button_Map [6] =
 {
    JMOUSE_BUTTON_1,
    JMOUSE_BUTTON_2,
-   JMOUSE_BUTTON_3
+   JMOUSE_BUTTON_3,
+   JMOUSE_BUTTON_4,
+   JMOUSE_BUTTON_5,
+   JMOUSE_BUTTON_6
 };
 
 /*}}}*/
@@ -87,22 +92,37 @@ void (*X_Close_Mouse_Hook)(void);
 #endif /* NOT USE_GPM_MOUSE */
 
 
-#if JED_HAS_TTY_MENUS
+#if JED_HAS_MENUS
 static int Hack_Hack = 1;
 static Window_Type *Non_Existent_Menu_Window = (Window_Type *)&Hack_Hack;
 #endif
 #define BUFFER_STATUS_LINE	1
 #define MENU_BAR_STATUS		2
 
-/* *ap and *bp must have one of values in set (1, 2, 4).  */
+/* returns button number from map */
+static unsigned char m2b (int m)
+{
+   switch (m)
+     {
+      default:
+      case 0x01: return 0;
+      case 0x02: return 1;
+      case 0x04: return 2;
+      case 0x08: return 3;
+      case 0x10: return 4;
+      case 0x20: return 5;
+     }
+}
+
+/* *ap and *bp must have one of values in set (1, 2, 4, 8, 16, 32, 64).  */
 void jed_map_mouse_buttons (int *ap, int *bp) /*{{{*/
 {
    unsigned char a, b;
    
-   a = (unsigned char) (*ap >> 1);
-   b = (unsigned char) (*bp >> 1);
+   a = m2b (*ap);
+   b = m2b (*bp);
    
-   if ((a > 2) || (b > 2)) return;
+   if ((a > 5) || (b > 5)) return;
    
    Mouse_Button_Map [a] = b << 1;
 }
@@ -162,8 +182,8 @@ static JMouse_Type *get_mouse_event (unsigned int *type) /*{{{*/
      return NULL;
    
    jm = Mouse_Queue + queue_pos;
-   /* 1, 2, 4 --> Mouse_Button_Map of 0, 1, 2 */
-   jm->button = Mouse_Button_Map [jm->button >> 1];
+   /* 1, 2, 4, 8,..., 32 --> Mouse_Button_Map of 0, 1, 2, ... 5 */
+   jm->button = Mouse_Button_Map [m2b(jm->button)];
    *type = jm->type;
    jm->type = 0;
    
@@ -220,7 +240,7 @@ static int switch_to_event_window (int x, int y, int *linep, int *colp, int *sta
    
    if (Use_This_Window != NULL)
      {
-#if JED_HAS_TTY_MENUS
+#if JED_HAS_MENUS
 	if (Use_This_Window == Non_Existent_Menu_Window)
 	  {
 	     *status = last_status = MENU_BAR_STATUS;
@@ -238,11 +258,9 @@ static int switch_to_event_window (int x, int y, int *linep, int *colp, int *sta
      }
    else
      {
-	int top, bot;
-
-#if JED_HAS_TTY_MENUS
+#if JED_HAS_MENUS
 	if (Jed_Menus_Active
-	    || ((y == 1) && (Top_Window_Row > 1)))
+	    || ((y == 1) && (Top_Window_SY != 0)))
 	  {
 	     Use_This_Window = Non_Existent_Menu_Window;
 	     *colp = x;
@@ -254,10 +272,12 @@ static int switch_to_event_window (int x, int y, int *linep, int *colp, int *sta
 	w = JWindow;
 	do
 	  {
-	     top = w->top;
-	     bot = top + w->rows;
+	     int sy, bot;
+
+	     sy = w->sy;
+	     bot = sy + w->rows + 1;
 	     
-	     if ((y >= top) && (y < bot))
+	     if ((y >= sy) && (y < bot))
 	       {
 		  Use_This_Window = w;
 		  break;
@@ -290,7 +310,7 @@ static int switch_to_event_window (int x, int y, int *linep, int *colp, int *sta
      }
    
    /* What line does y correspond ??? */
-   delta_y = y - (Use_This_Window->top - 1 + window_line ());
+   delta_y = y - (Use_This_Window->sy + window_line ());
    y = LineNum;
 #if JED_HAS_LINE_ATTRIBUTES
    if (delta_y == 0)
@@ -340,7 +360,7 @@ static int switch_to_event_window (int x, int y, int *linep, int *colp, int *sta
 #endif
 
    *linep = y + delta_y;
-   *colp = JWindow->column + x - 1;
+   *colp = JWindow->hscroll_column + x - 1;
    return 0;
 }
 
@@ -348,18 +368,18 @@ static int switch_to_event_window (int x, int y, int *linep, int *colp, int *sta
 
 static int do_function (SLang_Name_Type *fun, int line, int col, int button, int shift) /*{{{*/
 {
-   int ret = 1;
+   int ret;
+
+   if ((-1 == SLang_start_arg_list ())
+       || (-1 == SLang_push_integer (line))
+       || (-1 == SLang_push_integer (col))
+       || (-1 == SLang_push_integer (button))
+       || (-1 == SLang_push_integer (shift))
+       || (-1 == SLang_end_arg_list ()))
+     return 1;
    
-   SLang_push_integer (line);
-   SLang_push_integer (col);
-   SLang_push_integer (button);
-   SLang_push_integer (shift);
-   if (SLang_Error) return ret;
-   
-   SLexecute_function (fun);
-   
-   if ((SLang_Error == 0)
-       && SLang_pop_integer (&ret))
+   if ((-1 == SLexecute_function (fun))
+       || (-1 == SLang_pop_integer (&ret)))
      ret = 1;
    
    return ret;
@@ -380,9 +400,10 @@ static void switch_to_window (Window_Type *w) /*{{{*/
 static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
 {
    int button, shift;
-   int linenum, column, is_status;
+   int linenum = -1, column, is_status;
    int ret;
    SLang_Name_Type *fun, *default_fun;
+   Jed_Buffer_Hook_Type *h = CBuf->buffer_hooks;
 
    button = jmouse->button;
    
@@ -416,14 +437,15 @@ static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
      }
    
    default_fun = NULL;
-   
+   fun = NULL;
+
    switch (type)
      {
       case JMOUSE_DRAG:
 	if (is_status) fun = Jed_Mouse_Status_Drag_Hook;
 	else
 	  {
-	     fun = CBuf->mouse_drag_hook;
+	     if (h != NULL) fun = h->mouse_drag_hook;
 	     default_fun = Jed_Default_Mouse_Drag_Hook;
 	  }
 	break;
@@ -432,7 +454,7 @@ static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
 	if (is_status) fun = Jed_Mouse_Status_2Click_Hook;
 	else
 	  {
-	     fun = CBuf->mouse_2click_hook;
+	     if (h != NULL) fun = h->mouse_2click_hook;
 	     default_fun = Jed_Default_Mouse_2Click_Hook;
 	  }
 	break;
@@ -441,7 +463,7 @@ static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
 	if (is_status) fun = Jed_Mouse_Status_3Click_Hook;
 	else
 	  {
-	     fun = CBuf->mouse_3click_hook;
+	     if (h != NULL) fun = h->mouse_3click_hook;
 	     default_fun = Jed_Default_Mouse_3Click_Hook;
 	  }
 	break;
@@ -451,7 +473,7 @@ static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
 	if (is_status) fun = Jed_Mouse_Status_Down_Hook;
 	else
 	  {
-	     fun = CBuf->mouse_down_hook;
+	     if (h != NULL) fun = h->mouse_down_hook;
 	     default_fun = Jed_Default_Mouse_Down_Hook;
 	  }
 	break;
@@ -462,7 +484,7 @@ static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
 	if (is_status) fun = Jed_Mouse_Status_Up_Hook;
 	else
 	  {
-	     fun = CBuf->mouse_up_hook;
+	     if (h != NULL) fun = h->mouse_up_hook;
 	     default_fun = Jed_Default_Mouse_Up_Hook;
 	  }
 	break;
@@ -478,7 +500,7 @@ static int do_mouse_cmd (JMouse_Type *jmouse, unsigned int type) /*{{{*/
      shift = 2;
    else shift = 0;
 
-#if JED_HAS_TTY_MENUS
+#if JED_HAS_MENUS
    if (is_status == MENU_BAR_STATUS)
      return jed_menu_handle_mouse (type, column, linenum, button, shift);
 #endif
@@ -570,3 +592,5 @@ void jed_set_default_mouse_hook (char *hook_name, char *function) /*{{{*/
 }
 
 /*}}}*/
+
+#endif				       /* HAS_MOUSE */
