@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -45,7 +45,7 @@
 
 #include "dos_os2.c"
 
-extern char *get_cwd (void);
+/* extern char *get_cwd (void); */
 void delay (int time);
 
 void set_kbd (void);
@@ -356,7 +356,7 @@ int get_term_dimensions(int *w, int *h)
 #endif
 
 #if defined(_MSC_VER) || defined (__EMX__) || defined(__WATCOMC__)
-int sys_chmod(char *file, int what, int *mode, short *uid, short *gid)
+int sys_chmod(SLFUTURE_CONST char *file, int what, mode_t *mode, uid_t *uid, gid_t *gid)
 {
    struct stat buf;
    int m;
@@ -392,7 +392,7 @@ int sys_chmod(char *file, int what, int *mode, short *uid, short *gid)
 
 #else
 
-int sys_chmod(char *file, int what, int *mode, short *dum1, short *dum2)
+int sys_chmod(char *file, int what, mode_t *mode, uid_t *dum1, gid_t *dum2)
 {
    FILESTATUS3 fileInfo;
    USHORT Result;
@@ -451,7 +451,7 @@ int sys_findfirst(char *theFile)
    HDIR FindHandle;
    File_Attr = FILE_READONLY | FILE_DIRECTORY;
 
-   file = expand_filename(theFile);
+   file = jed_standardize_filename_static(theFile);
    f1 = f = extract_file(file);
    strcpy (Found_Dir, file);
 
