@@ -1,7 +1,9 @@
 /* -*- mode: C; mode: fold; -*- */
-/*
- *  Copyright (c) 1992, 1995 John E. Davis  (davis@space.mit.edu)
- *  All Rights Reserved.
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
+ * This file is part of JED editor library source.
+ *
+ * You may distribute this file under the terms the GNU General Public
+ * License.  See the file COPYING for more information.
  */
     
 #include "config.h"
@@ -25,7 +27,7 @@
 #define JED_MARK_TYPE 128
 
 Buffer *Paste_Buffer;
-Buffer *Rectangle_Buffer;
+static Buffer *Rectangle_Buffer;
 
 /* This is used by the narrow command so that multiple windows are 
  * handled properly. The assumption is that we are dealing with a canonical
@@ -48,16 +50,12 @@ static void touch_windows(void) /*{{{*/
 	     n = CBuf->marks->n;
 	     while ((l != NULL) && (l != JWindow->mark.line)) l = l->next, n++;
 	     if (l == NULL) 
-	       {
-		  JWindow->mark.line = CLine;
-		  JWindow->mark.point = Point;
-		  JWindow->mark.n = n - 1;
-	       }
+	       jed_init_mark (&JWindow->mark, 0);
 	     touch_window();
 	  }
 	JWindow = JWindow->next;
      }
-   pop_mark(&Number_Zero);
+   jed_pop_mark(0);
 }
 
 /*}}}*/
@@ -77,26 +75,68 @@ static int line_in_buffer(Line *line) /*{{{*/
 
 /*{{{ Spot/Mark Functions */
 
-static void init_mark (Mark *m, unsigned int flags) /*{{{*/
+void jed_copy_mark (Mark *dest, Mark *src)
+{
+   *dest = *src;
+   dest->next = NULL;
+}
+
+void jed_init_mark (Mark *m, unsigned int flags) /*{{{*/
 {
    m->line = CLine;
    m->point = Point;
    m->n = LineNum + CBuf->nup;
    m->flags = flags;
+   /* m->next = NULL; */
 }
 
 /*}}}*/
 
+void jed_init_mark_for_buffer (Mark *m, Buffer *b, unsigned int flags) /*{{{*/
+{
+   m->line = b->line;
+   m->point = b->point;
+   m->n = b->linenum + b->nup;
+   m->flags = flags;
+   /* m->next = NULL; */
+}
+
+/*}}}*/
+#if 0
+int jed_init_mark_for_line (Mark *m, Line *l, unsigned int flags) /*{{{*/
+{
+   Line *ll;
+   unsigned int n;
+   
+   n = 1;
+   ll = CBuf->beg;
+   while ((ll != NULL) && (ll != l))
+     {
+	ll = ll->next;
+	n++;
+     }
+   if (ll == NULL)
+     return -1;
+   
+   m->line = l;
+   m->point = 0;
+   m->n = n + CBuf->nup;
+   m->flags = flags;
+   /* m->next = NULL; */
+   return 0;
+}
+
+/*}}}*/
+#endif
 #if JED_HAS_SAVE_NARROW
 static Mark *create_mark (unsigned int flags) /*{{{*/
 {
    Mark *m;
    
-   if (NULL == (m = (Mark *) SLMALLOC(sizeof(Mark))))
-      {
-	  exit_error("create-mark: malloc error", 0);
-      }
-   init_mark (m, flags);
+   if (NULL == (m = (Mark *) jed_malloc0 (sizeof(Mark))))
+     exit_error("create-mark: malloc error", 0);
+   
+   jed_init_mark (m, flags);
    
    return m;
 }
@@ -109,19 +149,19 @@ int set_mark_cmd (void) /*{{{*/
    Mark *m;
    if (Repeat_Factor != NULL)
      {
-	while (CBuf->marks != NULL) pop_mark(&Number_Zero);
+	while (CBuf->marks != NULL) jed_pop_mark(0);
 	Repeat_Factor = NULL;
 	return(1);
      }
    
    if (CBuf->marks == NULL)
      {
-	push_mark ();
+	jed_push_mark ();
      }
 
    m = CBuf->marks;
    
-   init_mark (m, m->flags);
+   jed_init_mark (m, m->flags);
    
    if ((m->flags & VISIBLE_MARK) == 0) 
      {
@@ -146,24 +186,21 @@ static Mark *allocate_mark (Jed_Mark_Array_Type **ap, Mark **mp, unsigned int fl
    if ((a == NULL)
        || (a->num_marks == JED_MAX_MARK_ARRAY_SIZE))
      {
-	if (NULL == (b = (Jed_Mark_Array_Type *) SLMALLOC(sizeof(Jed_Mark_Array_Type))))
+	if (NULL == (b = (Jed_Mark_Array_Type *) jed_malloc0 (sizeof(Jed_Mark_Array_Type))))
 	  {
 	     exit_error("allocate_mark: malloc error", 0);
 	  }
-	memset ((char *) b, 0, sizeof (Jed_Mark_Array_Type));
 	b->next = a;
 	a = b;
 	*ap = a;
      }
    
    m = &a->marks[a->num_marks];
+   jed_init_mark (m, flags);
+
    a->num_marks += 1;
-   
    m->next = *mp;
    *mp = m;
-   
-   init_mark (m, flags);
-   
    return m;
 }
 
@@ -183,7 +220,7 @@ static void deallocate_mark (Jed_Mark_Array_Type **ap, Mark **mp) /*{{{*/
    if (a->num_marks == 0)
      {
 	*ap = a->next;
-	SLFREE (a);
+	SLfree ((char *)a);
      }
 }
 
@@ -197,7 +234,7 @@ int push_spot() /*{{{*/
 
 /*}}}*/
 
-int push_mark() /*{{{*/
+int jed_push_mark() /*{{{*/
 {
    (void) allocate_mark (&CBuf->mark_array, &CBuf->marks, 0);
    return 1;
@@ -205,7 +242,7 @@ int push_mark() /*{{{*/
 
 /*}}}*/
 
-int goto_mark(Mark *m) /*{{{*/
+int jed_goto_mark(Mark *m) /*{{{*/
 {
    Line *l;
    int ret = -1;
@@ -227,15 +264,15 @@ int goto_mark(Mark *m) /*{{{*/
 
 /*}}}*/
 
-int pop_mark (int *go) /*{{{*/
+int jed_pop_mark (int go) /*{{{*/
 {
    Mark *m;
    
    m = CBuf->marks;
    if (m == NULL) return(0);
 
-   if (*go) goto_mark(m);
-   if (m->flags & VISIBLE_MARK)
+   if (go) jed_goto_mark(m);
+   if (m->flags & VISIBLE_MARK_MASK)
      {
 	CBuf->vis_marks--;
 	/* touch screen since region may be highlighted */
@@ -265,7 +302,7 @@ static int pop_spot_go (int go) /*{{{*/
    m = CBuf->spots;
    if (m == NULL) return(0);
    
-   if (go) goto_mark (m);
+   if (go) jed_goto_mark (m);
 
    deallocate_mark (&CBuf->spot_array, &CBuf->spots);
    return 1;
@@ -293,7 +330,7 @@ int exchange_point_mark(void) /*{{{*/
    save_line = CLine;
    save_n = LineNum + CBuf->nup;
    
-   goto_mark (m);
+   jed_goto_mark (m);
    
    m->point = save_point; m->line = save_line; m->n = save_n;
    return(1);
@@ -317,7 +354,7 @@ int check_region(int *push) /*{{{*/
 
    if (CBuf->marks == NULL)
      {
-	msg_error("Set mark first.");
+	msg_error("No region defined");
 	return(0);
      }
 
@@ -355,13 +392,16 @@ static int widen_buffer_lines (Buffer *b) /*{{{*/
    push_spot();
    eob();
    if ((n->end != NULL)
-       && (!CLine->len || ('\n' != *(CLine->data + (CLine->len - 1)))))
+       && (0 == LINE_HAS_NEWLINE (CLine)))
      {
+	/* This is a hack to avoid messing with the lock file. Yuk. */
 	unsigned int flags = CBuf->flags;
-	ins('\n');
-	CBuf->flags = flags;
+	CBuf->flags &= ~(READ_ONLY);
+	CBuf->flags |= BUFFER_NON_LOCKING;
+	(void) _jed_ins_byte ('\n');
+	CBuf->flags = flags;  /* jed_set_buffer_flags (CBuf, flags); */
      }
-   
+
    pop_spot();
    
    if (n->end != NULL) n->end->prev = b->end;
@@ -378,8 +418,10 @@ static int widen_buffer_lines (Buffer *b) /*{{{*/
    b->nup -= n->nup;
    b->ndown -= n->ndown;
    b->narrow = n->next;
-   
-   SLFREE(n);
+
+   /* mark_undo_boundary (b); */
+
+   SLfree ((char *)n);
    switch_to_buffer(save);
    return(1);
 }
@@ -400,8 +442,11 @@ int widen_buffer (Buffer *b) /*{{{*/
      return widen_buffer_lines (b);
    
    flags = b->flags;
-   b->flags &= ~READ_ONLY;
-   
+   b->flags &= ~(READ_ONLY);
+
+   /* A hack to avoid locking file */
+   b->flags |= BUFFER_NON_LOCKING;
+
    save = CBuf;
    if (b != CBuf) switch_to_buffer (b);
    push_spot ();
@@ -412,23 +457,25 @@ int widen_buffer (Buffer *b) /*{{{*/
    
 #if JED_HAS_LINE_ATTRIBUTES
    line_flags = CLine->flags;
-   CLine->flags &= ~JED_LINE_READONLY;
+   CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
-   del ();
+   (void) jed_del_wchar ();
 #if JED_HAS_LINE_ATTRIBUTES
    CLine->flags = line_flags;
 #endif
    
    pop_spot ();
-   prevline (&Number_One);
+   (void) jed_up (1);
 #if JED_HAS_LINE_ATTRIBUTES
    line_flags = CLine->flags;
-   CLine->flags &= ~JED_LINE_READONLY;
+   CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
-   del ();
+   (void) jed_del_wchar ();
 #if JED_HAS_LINE_ATTRIBUTES
    CLine->flags = line_flags;
 #endif
+   /* mark_undo_boundary (b); */
+
    pop_spot ();
    if (save != CBuf) switch_to_buffer (save);
    
@@ -449,16 +496,20 @@ int narrow_to_region (void) /*{{{*/
      
    if (0 == check_region (&Number_One))/* spot pushed */
      return 0;
-   
+
+   /* unmark_undo_boundary (CBuf); */
+
    flags = CBuf->flags;
-   CBuf->flags &= ~READ_ONLY;
+   CBuf->flags &= ~(READ_ONLY);
+   /* A hack to avoid locking file */
+   CBuf->flags |= BUFFER_NON_LOCKING;
    
    push_spot ();
    
    line = CLine;
    pnt = Point;
    
-   pop_mark (&Number_One);
+   jed_pop_mark (1);
    /* Special case if region is empty */
    if ((CLine == line) && (pnt == Point))
      {
@@ -467,23 +518,23 @@ int narrow_to_region (void) /*{{{*/
 
 #if JED_HAS_LINE_ATTRIBUTES
 	line_flags = CLine->flags;
-	CLine->flags &= ~JED_LINE_READONLY;
+	CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
 	
-	newline ();
+	(void) jed_insert_newline ();
 	
 #if JED_HAS_LINE_ATTRIBUTES
-	CLine->prev->flags = line_flags;
+	if (CLine->prev != NULL) CLine->prev->flags = line_flags;
 #endif
 	
-	push_mark ();
+	jed_push_mark ();
 
-	newline ();
+	(void) jed_insert_newline ();
 	
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
 #endif
-	prevline (&Number_One);
+	(void) jed_up (1);
 
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
@@ -496,26 +547,26 @@ int narrow_to_region (void) /*{{{*/
      {
 #if JED_HAS_LINE_ATTRIBUTES
 	line_flags = CLine->flags;
-	CLine->flags &= ~JED_LINE_READONLY;
+	CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
-	newline ();
+	jed_insert_newline ();
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
-	CLine->prev->flags = line_flags;
+	if (CLine->prev != NULL) CLine->prev->flags = line_flags;
 #endif
 
-	push_mark ();
+	jed_push_mark ();
 	pop_spot ();
 	
 #if JED_HAS_LINE_ATTRIBUTES
 	line_flags = CLine->flags;
-	CLine->flags &= ~JED_LINE_READONLY;
+	CLine->flags &= ~JED_LINE_IS_READONLY;
 #endif
-	newline ();
+	jed_insert_newline ();
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
 #endif
-	prevline (&Number_One);
+	jed_up (1);
 
 #if JED_HAS_LINE_ATTRIBUTES
 	CLine->flags = line_flags;
@@ -526,22 +577,29 @@ int narrow_to_region (void) /*{{{*/
 	pop_spot ();
      }
    
+   /* mark_undo_boundary (CBuf); */
    CBuf->flags = flags;
    return 1;
 }
 
 /*}}}*/
 
-int widen () /*{{{*/
+int widen (void) /*{{{*/
 {
-   return widen_buffer_lines (CBuf);
+   if (CBuf->narrow == NULL)
+     return 0;
+
+   if (CBuf->narrow->is_region == 0)
+     return widen_buffer_lines (CBuf);
+
+   return widen_buffer (CBuf);
 }
 
 /*}}}*/
 
 int widen_region (void) /*{{{*/
 {
-   return widen_buffer (CBuf);
+   return widen ();
 }
 
 /*}}}*/
@@ -552,17 +610,16 @@ int narrow_to_lines () /*{{{*/
    Line *beg;
    Narrow_Type *nt;
    
-   if (NULL == (nt = (Narrow_Type *) SLMALLOC(sizeof(Narrow_Type))))
-     {
-	msg_error("Malloc Error during narrow.");
-	return(0);
-     }
+   if (NULL == (nt = (Narrow_Type *) jed_malloc0 (sizeof(Narrow_Type))))
+     return 0;
    
    if (!check_region(&Number_One)) return(0);       /* spot pushed */
 
+   /* unmark_undo_boundary (CBuf); */
+
    push_spot();
-   pop_mark(&Number_One);
-   push_mark();			       /* popped and used in touch_windows! */
+   jed_pop_mark(1);
+   jed_push_mark();			       /* popped and used in touch_windows! */
    beg = CLine;
    nt->nup = LineNum - 1;
    
@@ -617,9 +674,9 @@ int yank() /*{{{*/
 
 int copy_region_to_buffer(Buffer *b) /*{{{*/
 {
-   int first_point, last_point, n, tmpm;
+   int first_point, last_point, n;
    Line *first, *last;
-   Buffer *save_buffer;
+   Buffer *save_buf;
 
    if (b->flags & READ_ONLY) 
      {
@@ -631,7 +688,7 @@ int copy_region_to_buffer(Buffer *b) /*{{{*/
    last = CLine;
    last_point = Point;
 
-   tmpm = 1; pop_mark(&tmpm);
+   jed_pop_mark(1);
    if (b == CBuf)
      {
 	msg_error("A buffer cannot be inserted upon itself.");
@@ -642,7 +699,7 @@ int copy_region_to_buffer(Buffer *b) /*{{{*/
    first = CLine;
    first_point = Point;
 
-   save_buffer = CBuf;
+   save_buf = CBuf;
    switch_to_buffer(b);
 
    /* go through standard routines for undo comapatability */
@@ -650,23 +707,28 @@ int copy_region_to_buffer(Buffer *b) /*{{{*/
    if (first == last)
      {
 	n = last_point - first_point;
-	if (save_buffer == MiniBuffer)
+	if (save_buf == MiniBuffer)
 	  {
-	     ins_chars(first->data + first_point, n);
+	     (void) jed_insert_nbytes (first->data + first_point, n);
 	  }
-	else quick_insert(first->data + first_point, n);
+	else (void) jed_quick_insert (first->data + first_point, n);
      }
    else 
      {
 	n = first->len - first_point;
-	quick_insert(first->data + first_point, n);
+	if (-1 == jed_quick_insert(first->data + first_point, n))
+	  goto the_return;
+
 	while (first = first->next, first != last)
 	  {
-	     quick_insert(first->data, first->len);
+	     if (-1 == jed_quick_insert (first->data, first->len))
+	       goto the_return;
 	  }
-	quick_insert(first->data, last_point);
+	(void) jed_quick_insert(first->data, last_point);
      }
-   switch_to_buffer(save_buffer);
+
+   the_return:
+   switch_to_buffer(save_buf);
    pop_spot();
    return(1);
 }
@@ -677,8 +739,7 @@ int copy_to_pastebuffer() /*{{{*/
 {
    /* delete paste buffer */
    if (Paste_Buffer != NULL) delete_buffer(Paste_Buffer);
-   Paste_Buffer = make_buffer();
-   strcpy(Paste_Buffer->name, " <paste>");
+   Paste_Buffer = make_buffer (" <paste>", NULL, NULL);
 
    copy_region_to_buffer(Paste_Buffer);
    return(0);
@@ -686,70 +747,103 @@ int copy_to_pastebuffer() /*{{{*/
 
 /*}}}*/
 
+static int jed_check_readonly_region (void)
+{
+   int beg_point, end_point;
+   Line *beg, *end;
+
+   if (CBuf->flags & READ_ONLY) 
+     { 
+	msg_error(Read_Only_Error);
+	return -1;
+     }
+
+   if (!check_region(&Number_Zero))
+     return -1;
+
+   end = CLine; end_point = Point;
+   (void) exchange_point_mark ();
+   beg = CLine; beg_point = Point;
+   (void) exchange_point_mark ();
+   
+   while (1)
+     {
+	if (beg->flags & JED_LINE_IS_READONLY) 
+	  {
+	     if ((beg == end) 
+		 && (end_point == 0)
+		 && (beg_point == 0))
+	       return 0;
+
+	     msg_error (Line_Read_Only_Error);
+	     return -1;
+	  }
+	if (beg == end)
+	  break;
+
+	beg = beg->next;
+     }
+   return 0;
+}
+
+   
 int delete_region (void) /*{{{*/
 {
-   int beg_point, tmpm, end_point, n;
+   int beg_point, end_point;
    Line *beg, *end;
    
-   CHECK_READ_ONLY
-    if (!check_region(&Number_Zero)) return(0);
+   if (0 != jed_check_readonly_region ())
+     return -1;
 
    /* make this go through standard ins/del routines to ensure undo */
    
    end = CLine; end_point = Point;
    push_spot();
-   tmpm = 1; pop_mark(&tmpm);
+   jed_pop_mark(1);
    beg = CLine; beg_point = Point;
    pop_spot();
    
-   Point = 0;
-   
    if (end != beg) 
      {
-	deln(&end_point);
-	
+	bol ();
+   
+	if (-1 == jed_del_nbytes (end_point))
+	  return -1;
+
 	/* go back because we do not want to mess with Line structures
 	   changing on us --- shouldn't happen anyway */
 	
-	while (CLine = CLine->prev, LineNum--, CLine != beg)
+	while (jed_up (1) && (CLine != beg))
 	  {
-	     eol(); 
-	     n = Point + 1;
-	     Point = 0;
-	     generic_deln(&n);
+	     bol ();
+	     if (-1 == jed_del_through_eol ())
+	       return -1;
 	  }
-	eol();
-	n = Point - beg_point + 1;  /* the \n char */
-	Point = beg_point;
+	end_point = CLine->len;	       /* include \n */
      }
-   else 
-     {
-	Point = beg_point;
-	n = end_point - Point;
-     }
-   
-   generic_deln(&n);
-   return(1);
+
+   jed_set_point (beg_point);
+   (void) jed_generic_del_nbytes (end_point - Point);
+   return 1;
 }
 
 /*}}}*/
 
 int kill_region() /*{{{*/
 {
-   int tmpm = 1;
-   
-   CHECK_READ_ONLY
+   if (-1 == jed_check_readonly_region ())
+     return -1;
 
    /* need two marks for this one */
    push_spot();
-   if (!pop_mark(&tmpm))
+   if (!jed_pop_mark(1))
      {
 	check_region(&Number_Zero);
 	pop_spot();
 	return(0);
      }
-   push_mark();
-   push_mark();
+   jed_push_mark();
+   jed_push_mark();
    pop_spot();
 
    copy_to_pastebuffer();
@@ -770,7 +864,11 @@ int insert_rectangle() /*{{{*/
    Line *rline;
 
    CHECK_READ_ONLY
-   if (Rectangle_Buffer == NULL) return(0);
+   if (0 == buffer_exists (Rectangle_Buffer))
+     {
+	Rectangle_Buffer = NULL;
+	return 0;
+     }
 
    Suspend_Screen_Update = 1;
    c1 = calculate_column();
@@ -778,18 +876,15 @@ int insert_rectangle() /*{{{*/
    if (rline != NULL) while (1)
      {
 	goto_column(&c1);
-	quick_insert(rline->data, rline->len);
+	if (-1 == jed_quick_insert (rline->data, rline->len))
+	  return -1;
+
 	rline = rline->next;
 	if (rline == NULL) break;
-	if (CLine->next == NULL)
+	if (0 == jed_down (1))
 	  {
 	     eol();
-	     newline();
-	  }
-	else 
-	  {
-	     CLine = CLine->next;
-	     LineNum++;
+	     jed_insert_newline();
 	  }
      }
    return(1);
@@ -807,7 +902,7 @@ int open_rectangle() /*{{{*/
 
    c1 = calculate_column();
    save_line = CLine;
-   tmpm = 1; pop_mark(&tmpm);
+   tmpm = 1; jed_pop_mark(1);
    c2 = calculate_column();
    n = c2 - c1;
    if (n < 0)
@@ -820,10 +915,10 @@ int open_rectangle() /*{{{*/
    while(1)
      {
 	goto_column(&c1);
-	ins_char_n_times(' ', n);
+	if (-1 == jed_insert_wchar_n_times(' ', n))
+	  break;
 	if (CLine == save_line) break;
-	CLine = CLine->next;
-	LineNum++;
+	(void) jed_down (1);
      }
    pop_spot();
 
@@ -832,156 +927,99 @@ int open_rectangle() /*{{{*/
 
 /*}}}*/
 
-/* rectangle commands */
-int copy_rectangle() /*{{{*/
+/* MB Safe */
+static int copy_or_kill_rectangle (int kill) /*{{{*/
 {
-    Line *save_line, *line, *beg;
-    int c1, c2, dc, tmp, tmpm, dc_malloc;
-    unsigned char *p1, *p2, *data;
+   Line *save_line, *line, *beg;
+   int c1, c2, dc, tmp;
    
-    if (!check_region(&Number_One)) return(0);       /* spot pushed */
-    /* delete Rectangle buffer */
-    if (Rectangle_Buffer != NULL) delete_buffer(Rectangle_Buffer);
-
-    Rectangle_Buffer = make_buffer();
-    strcpy(Rectangle_Buffer->name, " <rect>");
-    c2 = calculate_column();
-    save_line = CLine;
-
-    tmpm = 1; pop_mark(&tmpm);
-    c1 = calculate_column();
-    if (c1 == c2)
-      {
-	 msg_error(Rect_Error);
-	 pop_spot();
-	 return(0);
-      }
-    if (c1 > c2)
-      {
-	 tmp = c1;
-	 c1 = c2;
-	 c2 = tmp;
-	 goto_column(&c1);
-      }
-
-    /* go through the region copying rectanglar blocks to Rectanglebuffer */
-    dc = c2 - c1;
-   if (dc == 1) dc_malloc = 2; else dc_malloc = dc;
-    line = beg = make_line1(dc_malloc);
-   
-    beg->prev = NULL;
-    while (1)
-      {
-	 data = line->data;
-	 /* p1 = data;
-	 p2 = data + dc;
-	 while (p1 < p2) *p1++ = ' '; */
-	 SLMEMSET ((char *) data, ' ', dc);
-	 
-	 line->len = dc;
-
-	 if (c1 == goto_column1(&c1))
-	   {
-	      p1 = CLine->data + Point;
-	      (void) goto_column1(&c2);
-	      p2 = CLine->data + Point;
-
-	      /* while(p1 < p2) *data++ = *p1++; */
-	      SLMEMCPY((char *) data, (char *) p1, (int) (p2 - p1));
-	   }
-	 if (CLine == save_line) break;
-	 CLine = CLine->next;
-	 LineNum++;
-
-	 line->next = make_line1(dc_malloc);
-	 line->next->prev = line;
-	 line = line->next;
-      }
-
-    line->next = NULL;
-
-    Rectangle_Buffer->line = Rectangle_Buffer->beg = beg;
-    Rectangle_Buffer->end = line;
-    Rectangle_Buffer->point = 0;
-
-    pop_spot();
-    return(0);
-}
-
-/*}}}*/
-
-int kill_rectangle() /*{{{*/
-{
-    Line *save_line, *line, *beg;
-   int c1, c2, dc, tmp, n, tmpm, dc_malloc;
-   unsigned char *p1, *p2, *data;
-
-   CHECK_READ_ONLY
-   if (!check_region(&Number_One)) return(0);
-
+   if (!check_region(&Number_One)) return(0);       /* spot pushed */
    /* delete Rectangle buffer */
    if (Rectangle_Buffer != NULL) delete_buffer(Rectangle_Buffer);
-
-   Rectangle_Buffer = make_buffer();
-   strcpy(Rectangle_Buffer->name, " <rect>");
+   
+   Rectangle_Buffer = make_buffer (" <rect>", NULL, NULL);
    c2 = calculate_column();
    save_line = CLine;
+   
+   jed_pop_mark(1);
+   c1 = calculate_column();
+   if (c1 == c2)
+     {
+	msg_error(Rect_Error);
+	pop_spot();
+	return(0);
+     }
+   if (c1 > c2)
+     {
+	tmp = c1;
+	c1 = c2;
+	c2 = tmp;
+	goto_column(&c1);
+     }
+   
+   dc = c2 - c1;
+   
+   /* go through the region copying rectanglar blocks to Rectanglebuffer */
 
-   tmpm = 1; pop_mark(&tmpm);
-    c1 = calculate_column();
-    if (c1 == c2)
-      {
-	 msg_error(Rect_Error);
-	 pop_spot();
-	 return(0);
-      }
-    if (c1 > c2)
-      {
-	 tmp = c1;
-	 c1 = c2;
-	 c2 = tmp;
-	 goto_column(&c1);
-      }
+   line = beg = NULL;
+   while (1)
+     {
+	int col2, col1;
+	unsigned int len, len1;
+	unsigned int nspaces;
+	unsigned int nbytes;
+	unsigned char *p1, *p2;
 
-   Suspend_Screen_Update = 1;
-    /* go through the region copying rectanglar blocks to Rectanglebuffer */
-    dc = c2 - c1;
-   if (dc == 1) dc_malloc = 2; else dc_malloc = dc;
-   		       /* length of 1 not usable here.
-			* See makeline1 to see why 
-			*/
-    line = beg = make_line1(dc_malloc);
-    beg->prev = NULL;
-    while (1)
-      {
-	 data = line->data;
-	 
-	 /* p2 = data + dc; p1 = data;
-	 while (p1 < p2) *p1++ = ' '; */
-	 SLMEMSET((char *) data, ' ', dc);
-	 line->len = dc;
+	col1 = goto_column1 (&c1);
+	p1 = CLine->data + Point;
+
+	if (col1 == c1)
+	  {
+	     col2 = goto_column1(&c2);
+	     p2 = CLine->data + Point;
+	  }
+	else
+	  {
+	     col2 = col1;
+	     p2 = p1;
+	  }
+	
+	nspaces = (unsigned int) (dc - (col2 - col1));
+	nbytes = (unsigned int) (p2 - p1);
+	len1 = len = nbytes + nspaces;
+	
+	/* Need to allocate at least 2 bytes, since a single byte is a 
+	 * signature of a line with a single character whose value is a 
+	 * newline character.  See make_line1 for details.
+	 */
+	if (len <= 1)
+	  len1++;
+
+	if (beg == NULL)
+	  {
+	     beg = line = make_line1 (len1);
+	     beg->prev = NULL;
+	  }
+	else
+	  {
+	     line->next = make_line1 (len1);
+	     line->next->prev = line;
+	     line = line->next;
+	  }
+	line->len = len;
+
+	memcpy ((char *) line->data, (char *) p1, nbytes);
+	memset ((char *) line->data + nbytes, ' ', nspaces);
+
+	if (kill && (nbytes != 0))
+	  {
+	     jed_position_point (p1);
+	     if (-1 == jed_del_nbytes (nbytes))
+	       break;
+	  }
 
-	 if (c1 == goto_column1(&c1))
-	   {
-	      p1 = CLine->data + Point;
-	      (void) goto_column1(&c2);
-	      p2 = CLine->data + Point;
-	      Point = (int) (p1 - CLine->data);
-	      n = (int) (p2 - p1);
-	      SLMEMCPY((char *) data, (char *) p1, n);
-	      deln (&n);
-	      /* while(n-- > 0)
-	        {
-		   *data++ = *p1;
-		   del();
-		} */
-	   }
 	 if (CLine == save_line) break;
-	 CLine = CLine->next;  LineNum++;
-
-	 line->next = make_line1(dc_malloc);
-	 line->next->prev = line;
-	 line = line->next;
+	 (void) jed_down(1);
       }
 
     line->next = NULL;
@@ -996,46 +1034,56 @@ int kill_rectangle() /*{{{*/
 
 /*}}}*/
 
-int blank_rectangle() /*{{{*/
+int copy_rectangle (void)
 {
-   int c1, n, c2, pnt, tmpm;
+   return copy_or_kill_rectangle (0);
+}
+
+int kill_rectangle (void)
+{
+   return copy_or_kill_rectangle (1);
+}
+
+int blank_rectangle (void) /*{{{*/
+{
+   int c1, c2;
    Line *save_line;
-   int nn;
-   
+   int ncols;
 
    CHECK_READ_ONLY
    if (!check_region(&Number_One)) return(0); /* push_spot(); performed */
 
    c1 = calculate_column();
    save_line = CLine;
-   tmpm = 1; pop_mark(&tmpm);
+   jed_pop_mark(1);
    c2 = calculate_column();
-   n = c2 - c1;
-   if (n < 0)
+   if (c1 > c2)
      {
-	n = -n;
+	int tmp = c1;
 	c1 = c2;
+	c2 = tmp;
      }
-   
+   ncols = c2 - c1;
+
    Suspend_Screen_Update = 1;
    while(1)
      {
-	goto_column(&c1);
+	int pnt;
+
+	goto_column (&c2);
 	pnt = Point;
-	eol();
-	nn = Point - pnt;
-	if (nn > n) nn = n;
-	Point = pnt;
-	
-	deln(&nn);
-	ins_char_n_times( ' ', nn);
+	goto_column (&c1);
 	
+	if (-1 == jed_del_nbytes (pnt - Point))
+	  break;
+
+	if (-1 == jed_insert_wchar_n_times( ' ', ncols))
+	  break;
+
 	if (CLine == save_line) break;
-	CLine = CLine->next;
-	LineNum++;
+	(void) jed_down (1);
      }
    pop_spot();
-   /* mark_buffer_modified(&Number_One); */
    return(1);
 }
 
@@ -1054,11 +1102,14 @@ typedef struct /*{{{*/
 /*}}}*/
 User_Mark_Type;
 
-static void free_user_mark (User_Mark_Type *um) /*{{{*/
+static void free_user_mark (SLtype type, VOID_STAR um_alias) /*{{{*/
 {
    Mark *m, *m1;
    Buffer *b;
-
+   User_Mark_Type *um;
+   
+   (void) type;
+   um = (User_Mark_Type *) um_alias;
    m1 = &um->m;
 
    /* The mark is only valid if the buffer that it was created for still
@@ -1081,7 +1132,7 @@ static void free_user_mark (User_Mark_Type *um) /*{{{*/
 	  }
      }
    
-   SLFREE (um);
+   SLfree ((char *)um);
 }
 
 /*}}}*/
@@ -1111,12 +1162,36 @@ static int mark_valid (Mark *m) /*{{{*/
 
 /*}}}*/
 
-int jed_move_user_object_mark (SLuser_Object_Type *uo) /*{{{*/
+static SLang_MMT_Type *pop_valid_user_mark (User_Mark_Type **ump)
+{   
+   SLang_MMT_Type *mmt;
+   User_Mark_Type *um;
+
+   *ump = NULL;
+
+   if (NULL == (mmt = SLang_pop_mmt (JED_MARK_TYPE)))
+     return NULL;
+
+   um = (User_Mark_Type *) SLang_object_from_mmt (mmt);
+   
+   if (0 == mark_valid (&um->m))
+     {
+	SLang_free_mmt (mmt);
+	return NULL;
+     }
+   
+   *ump = um;
+   return mmt;
+}
+
+   
+   
+int jed_move_user_object_mark (SLang_MMT_Type *mmt) /*{{{*/
 {
    User_Mark_Type *um;
    Mark *m;
    
-   um = (User_Mark_Type *) uo->obj;
+   um = (User_Mark_Type *) SLang_object_from_mmt (mmt);
    m = &um->m;
    
    if (!mark_valid (m)) return 0;
@@ -1138,31 +1213,31 @@ int jed_move_user_object_mark (SLuser_Object_Type *uo) /*{{{*/
 
 void move_user_mark (void) /*{{{*/
 {
-   SLuser_Object_Type *uo;
+   SLang_MMT_Type *mmt;
    
-   if ((uo = SLang_pop_user_object (JED_MARK_TYPE)) == NULL) return;
-   (void) jed_move_user_object_mark (uo);
-   SLang_free_user_object (uo);
+   if (NULL == (mmt = SLang_pop_mmt (JED_MARK_TYPE)))
+     return;
+
+   (void) jed_move_user_object_mark (mmt);
+   SLang_free_mmt (mmt);
 }
 
 /*}}}*/
 
 static int x_user_mark_fun (int (*xfun)(Mark *)) /*{{{*/
 {
-   SLuser_Object_Type *uo;
+   SLang_MMT_Type *mmt;
    User_Mark_Type *um;
    int ret = -1;
-   
-   if ((uo = SLang_pop_user_object (JED_MARK_TYPE)) == NULL) return -1;
-   um = (User_Mark_Type *) uo->obj;
-   
-   if (mark_valid (&um->m))
-     {
-	if (CBuf != um->b) msg_error ("Mark not in buffer.");
-	else
-	  ret = (*xfun) (&um->m);
-     }
-   SLang_free_user_object (uo);
+
+   if (NULL == (mmt = pop_valid_user_mark (&um)))
+     return -1;
+
+   if (CBuf != um->b) msg_error ("Mark not in buffer.");
+   else
+     ret = (*xfun) (&um->m);
+
+   SLang_free_mmt (mmt);
    return ret;
 }
 
@@ -1182,101 +1257,264 @@ int jed_is_user_mark_in_narrow (void)
 
 void goto_user_mark (void)
 {
-   (void) x_user_mark_fun (goto_mark);
+   (void) x_user_mark_fun (jed_goto_mark);
 }
 
-SLuser_Object_Type *jed_make_user_object_mark (int main_type) /*{{{*/
+SLang_MMT_Type *jed_make_user_object_mark (void) /*{{{*/
 {
    User_Mark_Type *um;
-   SLuser_Object_Type *uo;
+   SLang_MMT_Type *mmt;
    Mark *m;
    
-   uo = SLang_create_user_object (JED_MARK_TYPE);
-   if (uo == NULL) return NULL;
-   uo->main_type = main_type;
-   
-   if (NULL == (um = (User_Mark_Type *) SLMALLOC (sizeof(User_Mark_Type))))
+   if (NULL == (um = (User_Mark_Type *) jed_malloc0 (sizeof(User_Mark_Type))))
+     return NULL;
+     
+   if (NULL == (mmt = SLang_create_mmt (JED_MARK_TYPE, (VOID_STAR) um)))
      {
-	SLang_Error = SL_MALLOC_ERROR;
-	SLFREE (uo);
+	SLfree ((char *) um);
 	return NULL;
      }
+
    m = &um->m;
    
-   init_mark (m, 0);
+   jed_init_mark (m, 0);
    
    m->next = CBuf->user_marks;
    
    CBuf->user_marks = m;
    
-   um->b = CBuf;   
-   uo->obj = (long *) um;
-   return uo;
+   um->b = CBuf;
+
+   return mmt;
 }
 
 /*}}}*/
 
-void jed_free_user_object_mark (SLuser_Object_Type *uo) /*{{{*/
-{
-   /* let slang handle this. */
-   uo->main_type = SLANG_DATA;
-   SLang_free_user_object (uo);
-}
-
-/*}}}*/
 
 void create_user_mark (void) /*{{{*/
 {
-   SLuser_Object_Type *uo = jed_make_user_object_mark (SLANG_DATA);
-   if (uo != NULL) SLang_push_user_object (uo);
+   SLang_MMT_Type *mmt;
+   
+   if (NULL != (mmt = jed_make_user_object_mark ()))
+     {
+	if (-1 == SLang_push_mmt (mmt))
+	  SLang_free_mmt (mmt);
+     }
 }
 
 /*}}}*/
 
 char *user_mark_buffer (void) /*{{{*/
 {
-   SLuser_Object_Type *uo;
+   SLang_MMT_Type *mmt;
    User_Mark_Type *um;
-   char *s = "";
-   
-   if (NULL == (uo = SLang_pop_user_object (JED_MARK_TYPE))) return s;
-   um = (User_Mark_Type *) uo->obj;
-   
-   if (mark_valid (&um->m))
-     {
-	s = um->b->name;
-     }
-   
-   SLang_free_user_object (uo);
+   char *s;
+
+   if (NULL == (mmt = pop_valid_user_mark (&um)))
+     return "";
+
+   s = um->b->name;
+
+   SLang_free_mmt (mmt);
    return s;
 }
 
 /*}}}*/
 
+static int
+user_mark_bin_op_result (int op, SLtype a, SLtype b,
+			 SLtype *c)
+{
+   (void) a; (void) b;
+   switch (op)
+     {
+      default: 
+	return 0;
+	
+      case SLANG_GT:
+      case SLANG_GE:
+      case SLANG_LT:
+      case SLANG_LE:
+      case SLANG_EQ:
+      case SLANG_NE:
+	*c = SLANG_INT_TYPE;
+	break;
+     }
+   return 1;
+}
+
+
+static int
+user_mark_bin_op (int op,
+		  SLtype a_type, VOID_STAR ap, unsigned int na,
+		  SLtype b_type, VOID_STAR bp, unsigned int nb,
+		  VOID_STAR cp)
+{
+   int *ic;
+   unsigned int n, n_max;
+   unsigned int da, db;
+   SLang_MMT_Type **a, **b;
+
+   (void) a_type;
+   (void) b_type;
+
+   if (na == 1) da = 0; else da = 1;
+   if (nb == 1) db = 0; else db = 1;
+
+   if (na > nb) n_max = na; else n_max = nb;
+
+   a = (SLang_MMT_Type **) ap;
+   b = (SLang_MMT_Type **) bp;
+
+   ic = (int *) cp;
+
+   for (n = 0; n < n_max; n++)
+     {
+	User_Mark_Type *ua, *ub;
+	Buffer *ba, *bb;
+	int pa, pb;
+	unsigned int la, lb;
+
+	ub = NULL;
+	ba = bb = NULL;
+	pa = pb = 0;
+	la = lb = 0;
+	
+	if ((*a != NULL)
+	    && (NULL != (ua = (User_Mark_Type *) SLang_object_from_mmt (*a))))
+	  {
+	     la = ua->m.n;
+	     pa = ua->m.point;
+	     ba = ua->b;
+	  }
+
+	if ((*b != NULL)
+	    && (NULL != (ub = (User_Mark_Type *) SLang_object_from_mmt (*b))))
+	  {
+	     lb = ub->m.n;
+	     pb = ub->m.point;
+	     bb = ub->b;
+	  }
+
+	if ((ba == NULL) || (bb == NULL))
+	  {
+	     ic[n] = 0;
+	     a += da;
+	     b += db;
+	     continue;
+	  }
+	
+	switch (op)
+	  {
+	   case SLANG_GT:
+	     ic[n] = ((la > lb)
+		      || ((la == lb) && (pa > pb)));
+	     break;
+	     
+	   case SLANG_GE:
+	     ic[n] = ((la > lb)
+		      || ((la == lb) && (pa >= pb)));
+	     break;
+	     
+	   case SLANG_LT:
+	     ic[n] = ((la < lb)
+		      || ((la == lb) && (pa < pb)));
+	     break;
+
+	   case SLANG_LE:
+	     ic[n] = ((la < lb)
+		      || ((la == lb) && (pa <= pb)));
+	     break;
+	     
+	   case SLANG_EQ:
+	     ic[n] = ((ba == bb) && (la == lb) && (pa == pb));
+	     break;
+
+	   case SLANG_NE:
+	     ic[n] = ((ba != bb) || (la != lb) || (pa != pb));
+	     break;
+	     
+	   default:
+	     return 0;
+	  }
+
+	a += da;
+	b += db;
+     }
+
+   return 1;
+}
+
+
+
 
 /*}}}*/
 
 #if JED_HAS_LINE_MARKS
 void jed_create_line_mark (int *color) /*{{{*/
 {
-   SLuser_Object_Type *uo = jed_make_user_object_mark (SLANG_DATA);
+   SLang_MMT_Type *mmt;
    User_Mark_Type *um;
    
-   if (uo != NULL) 
-     {
-	um = (User_Mark_Type *) uo->obj;
-	um->m.flags |= JED_LINE_MARK | (*color & MARK_COLOR_MASK);
-	SLang_push_user_object (uo);
-     }
+   if (NULL == (mmt = jed_make_user_object_mark ()))
+     return;
+
+   um = (User_Mark_Type *) SLang_object_from_mmt (mmt);
+   um->m.flags |= JED_LINE_MARK | (*color & MARK_COLOR_MASK);
+   if (-1 == SLang_push_mmt (mmt))
+     SLang_free_mmt (mmt);
 }
 
 /*}}}*/
 #endif
+
+#ifndef SLFUTURE_CONST
+# define SLFUTURE_CONST
+#endif
+
+static int user_mark_sget (SLtype type, SLFUTURE_CONST char *name)
+{
+   SLang_MMT_Type *mmt;
+   User_Mark_Type *um;
+   int status;
+   Buffer *buf;
+
+   (void) type;
    
+   if (NULL == (mmt = pop_valid_user_mark (&um)))
+     return -1;
+   
+   buf = um->b;
+
+   status = -1;
+   if (0 == strcmp (name, "buffer_name"))
+     status = SLang_push_string (buf->name);
+   else
+     SLang_verror (SL_NOT_IMPLEMENTED,
+		   "Mark_Type.%s is invalid", name);
+   
+   SLang_free_mmt (mmt);
+   return status;
+}
+
 
 int register_jed_classes (void) /*{{{*/
 {
-   return SLang_register_class (JED_MARK_TYPE, (FVOID_STAR) free_user_mark, NULL);
+   SLang_Class_Type *cl;
+   
+   cl = SLclass_allocate_class ("Mark_Type");
+   if (cl == NULL) return -1;
+   (void) SLclass_set_destroy_function (cl, free_user_mark);
+   
+   (void) SLclass_set_sget_function (cl, user_mark_sget);
+
+   if (-1 == SLclass_register_class (cl, JED_MARK_TYPE, sizeof (User_Mark_Type), SLANG_CLASS_TYPE_MMT))
+     return -1;
+   
+   if (-1 == SLclass_add_binary_op (JED_MARK_TYPE, JED_MARK_TYPE, user_mark_bin_op, user_mark_bin_op_result))
+     return -1;
+   
+   return 0;
 }
 
 /*}}}*/
@@ -1307,9 +1545,9 @@ static void restore_saved_narrow (void) /*{{{*/
    
    while (beg != NULL)
      {
-	goto_mark (beg);
-	push_mark ();
-	goto_mark (end);
+	jed_goto_mark (beg);
+	jed_push_mark ();
+	jed_goto_mark (end);
 	if (end->flags & NARROW_REGION_MARK)
 	  narrow_to_region ();
 	else narrow_to_lines ();
@@ -1328,24 +1566,24 @@ static void free_mark_chain (Mark *m) /*{{{*/
    while (m != NULL)
      {
 	next = m->next;
-	SLFREE (m);
+	SLfree ((char *)m);
 	m = next;
      }
 }
 
 /*}}}*/
-void jed_free_saved_narrow (void) /*{{{*/
+void jed_free_saved_narrow (Buffer *b) /*{{{*/
 {
    Jed_Save_Narrow_Type *save_narrow;
    
-   save_narrow = CBuf->save_narrow;
+   save_narrow = b->save_narrow;
    if (save_narrow == NULL) return;
    
-   CBuf->save_narrow = save_narrow->next;
+   b->save_narrow = save_narrow->next;
    
    free_mark_chain (save_narrow->beg);
    free_mark_chain (save_narrow->end);
-   SLFREE (save_narrow);
+   SLfree ((char *)save_narrow);
 }
 
 /*}}}*/
@@ -1353,7 +1591,7 @@ void jed_push_narrow (void) /*{{{*/
 {
    Jed_Save_Narrow_Type *save_narrow;
    
-   if (NULL == (save_narrow = (Jed_Save_Narrow_Type *) SLMALLOC (sizeof (Jed_Save_Narrow_Type))))
+   if (NULL == (save_narrow = (Jed_Save_Narrow_Type *) jed_malloc0 (sizeof (Jed_Save_Narrow_Type))))
      {
 	exit_error ("push_narrow: malloc error.", 0);
      }
@@ -1388,7 +1626,7 @@ void jed_push_narrow (void) /*{{{*/
 void jed_pop_narrow (void) /*{{{*/
 {
    restore_saved_narrow ();
-   jed_free_saved_narrow ();
+   jed_free_saved_narrow (CBuf);
 }
 
 /*}}}*/
@@ -1408,3 +1646,24 @@ int jed_count_narrows (void) /*{{{*/
 }
 
 /*}}}*/
+
+unsigned int jed_count_lines_in_region (void)
+{
+   Mark *m;
+   unsigned int n0, n1;
+
+   m = CBuf->marks;
+   if (m == NULL)
+     return 0;
+   
+   n0 = m->n - CBuf->nup;
+   n1 = LineNum;
+   
+   if (n0 > n1)
+     return 1 + (n0 - n1);
+   
+   return 1 + (n1 - n0);
+}
+
+	
+   
