@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -7,9 +7,9 @@
 
 extern int set_mark_cmd(void);
 extern int push_spot(void);
-extern int push_mark(void);
-extern int goto_mark(Mark *);
-extern int pop_mark(int *);
+extern int jed_push_mark (void);
+extern int jed_goto_mark(Mark *);
+extern int jed_pop_mark (int);
 extern int mark_spot(void);
 extern int pop_spot(void);
 extern int widen_buffer(Buffer *);
@@ -18,6 +18,11 @@ extern int widen_region (void);
 extern int narrow_to_region (void);
 extern int narrow_to_lines (void);
 extern void jed_widen_whole_buffer (Buffer *);
+extern void jed_init_mark (Mark *, unsigned int);
+extern void jed_init_mark_for_buffer (Mark *, Buffer *, unsigned int);
+extern int jed_init_mark_for_line (Mark *, Line *, unsigned int);
+extern void jed_copy_mark (Mark *dest, Mark *src);
+
 #if JED_HAS_SAVE_NARROW
 extern void jed_free_saved_narrow (Buffer *);
 extern void jed_push_narrow (void);
@@ -51,5 +56,6 @@ extern int register_jed_classes (void);
 extern void jed_create_line_mark (int *);
 #endif
 
+extern unsigned int jed_count_lines_in_region (void);
 extern Buffer *Paste_Buffer;
 
