@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -18,10 +18,18 @@
 # include <stdlib.h>
 #endif
 
+#define __USE_XOPEN_EXTENDED	       /* to get getpgid */
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
 
+#ifdef __WIN32__
+/* windows.h must be included before slang.h */
+# include <windows.h>
+#endif
+
+#include <slang.h>
+
 #include "jdmacros.h"
 
 #ifdef POSIX
@@ -34,13 +42,27 @@
 # ifdef VOID 
 #  undef VOID
 # endif
-# include <windows.h>
+# if !defined(__CYGWIN32__)
+#  define sleep Sleep
+# endif
 # undef _POSIX_SOURCE
 # include <sys/stat.h>
+# ifdef __BORLANDC__
+#  include <dir.h>
+# endif
+# if defined(__MINGW32__) || defined(__IBMC__)
+#  include <direct.h>
+# endif
+# ifdef __IBMC__
+#  if SLANG_VERSION < 10308
+#   define sleep(x) sys_pause(1000 * (x))
+#  else
+/* sleep added to slang 10308 */
+extern unsigned int sleep (unsigned int);
+#  endif
+# endif
 #endif
 
-#include <slang.h>
-
 #include "buffer.h"
 #include "sysdep.h"
 #include "display.h"
@@ -58,6 +80,13 @@ int (*X_Init_Term_Hook) (void);
 void (*X_Reset_Term_Hook) (void);
 
 int Ignore_User_Abort = 1;	       /* Abort char triggers S-Lang error */
+char *_Jed_Backspace_Key = "\x7F";
+
+void (*Jed_Sig_Exit_Fun) (void);
+
+#ifdef IBMPC_SYSTEM
+int Jed_Filename_Case_Sensitive = 0;
+#endif
 
 unsigned char KeyBoard_Xlate[256] = /*{{{*/
 {
@@ -140,15 +169,28 @@ int DEC_8Bit_Hack = 0;
 int my_getkey() /*{{{*/
 {
    char buf[10];
-   int i, imax;
+   int i;
+#if 0
+   int imax;
+#endif
    unsigned char ch;
-   
+   int eightbit_hack;
+
    if (Batch)
      {
 	fgets(buf, 9 ,stdin);
-	return (int) *buf;
+	return (int) (unsigned char)*buf;
      }
-   
+
+   /* Apparantly this hack causes problems when dealing with a unicode-aware
+    * terminal.
+    */
+
+   eightbit_hack = DEC_8Bit_Hack;
+#if JED_HAS_UTF8_SUPPORT
+   if (Jed_UTF8_Mode)
+     eightbit_hack = 0;
+#endif
    if (!Input_Buffer_Len)
      {
 	/* if (Batch) ch = (unsigned char) getc(stdin); else ch = sys_getkey(); */
@@ -156,9 +198,9 @@ int my_getkey() /*{{{*/
 	if (ch & 0x80)
 	  {
 	     i = (ch & 0x7F);
-	     if ((i < ' ') && DEC_8Bit_Hack)
+	     if ((i < ' ') && eightbit_hack)
 	       {
-		  i += DEC_8Bit_Hack;
+		  i += eightbit_hack;
 		  ungetkey((int *) &i);
 		  ch = 27;
 	       }
@@ -172,12 +214,12 @@ int my_getkey() /*{{{*/
      }
    
    ch = Input_Buffer[0];
-   if ((ch & 0x80) && ((Meta_Char != -1) || ((ch < 160) && DEC_8Bit_Hack)))
+   if ((ch & 0x80) && ((Meta_Char != -1) || ((ch < 160) && eightbit_hack)))
      {
 	ch = (ch & 0x7F);
-	if ((ch < ' ') && DEC_8Bit_Hack)
+	if ((ch < ' ') && eightbit_hack)
 	  {
-	     ch += DEC_8Bit_Hack;
+	     ch += eightbit_hack;
 	     i = 27;
 	  }
 	else i = Meta_Char;
@@ -189,22 +231,24 @@ int my_getkey() /*{{{*/
 #ifdef USING_INPUT_BUFFER
    USING_INPUT_BUFFER
 #endif
-     
-     Input_Buffer_Len--;
-   imax = Input_Buffer_Len;
-   
-   SLMEMCPY ((char *) Input_Buffer, (char *) (Input_Buffer + 1), imax);
+
+   Input_Buffer_Len--;
+#if 1
+   memmove (Input_Buffer, Input_Buffer+1, Input_Buffer_Len);
+#else
+   for (i = 0; i < imax; i++)
+     Input_Buffer[i] = Input_Buffer[i+1];
+#endif
+   /* SLMEMCPY ((char *) Input_Buffer, (char *) (Input_Buffer + 1), imax); */
    
 #ifdef DONE_WITH_INPUT_BUFFER
    DONE_WITH_INPUT_BUFFER
-#endif
-     
+#endif     
      return((int) ch);
 }
 
 /*}}}*/
 
-
 void ungetkey_string(char *s, int n) /*{{{*/
 {
    /* int i; */
@@ -270,6 +314,9 @@ int input_pending (int *tsecs) /*{{{*/
    /* FIXME!!  This should affect the macro buffer */
    if (Executing_Keyboard_Macro) return 1;
 
+   if (Jed_Sig_Exit_Fun != NULL)
+     (*Jed_Sig_Exit_Fun) ();
+
    if (Input_Buffer_Len) return Input_Buffer_Len;
    
    n = sys_input_pending (tsecs, 0);
@@ -355,11 +402,37 @@ char *slash2slash(char *dir) /*{{{*/
  * Note: If the file ends in a slash as in a/b/c/, then a pointer to
  * the END of the string is returned.
  */
-char *extract_file(char *file) /*{{{*/
+char *extract_file(SLFUTURE_CONST char *file) /*{{{*/
+{
+   SLFUTURE_CONST char *f;
+   
+   f = file + strlen(file);
+   while (f > file)
+     {
+	f--;
+	if (*f == SLASH_CHAR) return (char *)f + 1;
+     }
+   return (char *) file;
+}
+
+/*}}}*/
+
+/* given a canonical filename, return pointer to its name.  
+ * Note: If the file ends in a slash as in a/b/c/, then a pointer to
+ * the c/ is returned.
+ */
+char *jed_extract_file_or_dir (char *file) /*{{{*/
 {
    char *f;
    
    f = file + strlen(file);
+   if (f > file) 
+     {
+	f--;
+	if (*f == SLASH_CHAR)
+	  f--;
+     }
+
    while (f > file)
      {
 	f--;
@@ -369,166 +442,284 @@ char *extract_file(char *file) /*{{{*/
 }
 
 /*}}}*/
+#ifdef IBMPC_SYSTEM
+/* I do not know how to do this in a portable fashion.  
+ */
+static int pcsystem_getdisk (void)
+{
+#if defined(__MINGW32__) || defined(__IBMC__)
+   return _getdrive ();
+#else
+# if defined(__DJGPP__)
+   int d;  _dos_getdrive (&d); return d;
+# else
+#  if defined(__BORLANDC__)
+   return 1 + getdisk();
+#  else
+   /* This may not work */
+   char *cwd = jed_get_cwd ();
 
+   if ((cwd == NULL) || (*cwd == 0) || (cwd[1] != ':'))
+     cwd = "C:";
+
+   return 1 + (UPPER_CASE (*cwd) - 'A');
+#  endif			       /* __BORLANDC__ */
+# endif				       /* __DJGPP__ */
+#endif				       /* __MINGW32__ */
+}
+#endif				       /* IBMPC_SYSTEM */
 
 #ifndef VMS
-/* this routine returns a Static pointer which is considered volatile */
-char *expand_filename (char *file) /*{{{*/
+
+static void str_move_bytes (char *a, char *b)
 {
-   register char *p, ch;
-   char  *last, *p1;
-# if defined (IBMPC_SYSTEM)
-   static char work[JED_MAX_PATH_LEN];
-# else
-   static char work[2 * JED_MAX_PATH_LEN];
-# endif
-   char filebuf [JED_MAX_PATH_LEN];
-
-   /* Do not mess with the file passed in as an argument */
-   safe_strcpy (filebuf, file, sizeof (filebuf));
-   file = filebuf;
-
-   *work = 0;
-   /* the following combinations indicate non-relative path names:
-    *	"//"	path from the root dir
-    *	"~/"	path from the $HOME dir
-    * for dos, os2 only
-    *   "x:/"	path from "x:/" dir
-    *   "x:"	same as "x:/"
-    */
-   
-   p = slash2slash(file) + strlen(file);
-   
-   while (p > file)
-     {
-	ch = *p--;
-	if (ch == SLASH_CHAR )
-	  {
-	     if ( *p == SLASH_CHAR ) /* "//" combination */
-	       {
-# ifdef MSWINDOWS
-		  if (file == p)       /* With WFW network '\\host\filename' can be used */
-		    safe_strcpy(work, p, sizeof (work));
-		  else
-		    safe_strcpy(work, (p + 1), sizeof (work));
-# else
-#  ifdef __QNX__
-		    /* QNX pathnames look like //<node number>/rest_of_path 
-		     *  -- we need to leave on the <node number> 
-		     */
-		  if((*(p + 2) >= '1') && (*(p + 2) <= '9'))
-		    safe_strcpy(work, (p), sizeof (work));
-		  else
-		    safe_strcpy(work, (p + 1), sizeof (work));
-#  else
-		  safe_strcpy(work, (p + 1), sizeof (work));
-#  endif			       /* __QNX__ */
-# endif				       /* MSWINDOWS */
-		  file = work;
-		  break;
-	       }
-	     else if ((*p == '~')  /* "~/" combination */
-		      && (((p > file) && (*(p - 1) == SLASH_CHAR))
-			  || (p == file)))
-	       {
-		  if ( (p1 = getenv("HOME")) == NULL) p1 = "/";
-		  safe_strcpy( work, p1, sizeof (work) );
-		  /* remove trailing slash if any */
-		  p1 = slash2slash(work) + (strlen(work) - 1);
-		  if ( *p1 == SLASH_CHAR ) *p1 = '\0';
-		  safe_strcat (work, (p + 1), sizeof (work) );
-		  file = work;
-		  while (*file && (*file != SLASH_CHAR)) file++;
-		  break;
-	       }
-	  }
-# if defined (IBMPC_SYSTEM)		       /* DOS, OS/2 stuff */
-	else if (ch == ':' ) 	       /* "c:" or "c:/" combination */
-	  {
-	     safe_strcpy( work, p, sizeof (work) );
-	     file = (work+2);	       /* start file past the drive spec */
-	     p += 2;
-	     if ( *p != SLASH_CHAR ) 	       /* "c:" combination */
-	       {
-		  strcpy( file, "\\");
-		  safe_strcat (work, p, sizeof (work));
-	       }
-	     break;
-	  }
-# endif /* IBMPC_SYSTEM */
-     }
-   
-   if ( *work == '\0' )	       /* no special combinations */
-     {
-	if ( *file != SLASH_CHAR )
-	  {
-	     strcpy(work, get_cwd());      /* assume relative dir */
-	     slash2slash(work);
-	  }
-	safe_strcat(work, file, sizeof (work));
-	file = work;
-     }
-   
-   /* remove ../ and ./ stuff.  These combinations are only relevant when
-    * immediately preceed by a / character.  For simplicity in the algorithm
-    * several passes will be made.  In most cases, only one pass will be 
-    * required.
-    * 
-    * last is a pointer to the character following last / character where 
-    * ../ would backup to.  (This can also be at the beginning).
-    */
-   
-   p = last = file;
-   
-   while ((ch = *p++) != 0)
-     {
-	if (ch != SLASH_CHAR)
-	  continue;
-	
-	p1 = last;
-	last = p;
-	
-	/* A / seen.  Look for ./ or ../ */
-	if ((ch = *p) != '.')
-	  continue;
-	
-	p++;
-	ch = *p;
-	
-	if (ch != SLASH_CHAR)
-	  {
-	     if (ch != '.')
-	       continue;
-	     
-	     p++;
-	     ch = *p;
-	     if (ch != SLASH_CHAR)
-	       continue;
-	     
-	     /* We have a /../ combination.  p1 points to position that 
-	      * following characters should be copied to.
-	      */
-	     last = p1;
-	  }
-	/* else we have /./ combination.  Following characters should be 
-	 * copied to last position.
-	 */
-	
-	do
-	  {
-	     p++;
-	     ch = *p;
-	     *last++ = ch;
-	  }
-	while (ch != 0);
-
-	last = p = file;	       /* rescan */
-     }
-   
-   return work;
+   while (*b != 0)
+     *a++ = *b++;
+   *a = 0;
 }
 
+static char *strcat_malloc (char *a, char *b)
+{
+   char *c;
+   unsigned int lena = strlen (a);
+   unsigned int lenb = strlen (b);
+   
+   if (NULL == (c = SLmalloc (lena + lenb + 1)))
+     return NULL;
+   
+   strcpy (c, a);
+   strcpy (c+lena, b);
+   return c;
+}
+   
+/* file is assumed to be malloced here.  Its contents may be changed. */
+static int remove_slash_slash (char *file)
+{
+   char *f, *g, ch;
+   
+   g = f = file;
+   while ((ch = *f++) != 0)
+     {
+	*g++ = ch;
+	if (ch != SLASH_CHAR)
+	  continue;
+
+	while (*f == SLASH_CHAR)
+	  f++;
+     }
+   *g = 0;
+   return 0;
+}
+
+/* /foo/blah/../bar --> /foo/bar
+ * /foo/blah/./bar --> /foo/blah/bar
+ */
+static int remove_dot_dot_slashes (char *file)
+{
+   char *f = file;
+   
+   while (1)
+     {
+	char *f1;
+
+	while (*f && (*f != SLASH_CHAR))
+	  f++;
+	if (*f == 0)
+	  break;
+	f++;
+	if (*f != '.')
+	  continue;
+
+	/* foo/./bar -> foo/bar */
+	if (f[1] == SLASH_CHAR)
+	  {
+	     str_move_bytes (f, f+2);
+	     f--;
+	     continue;
+	  }
+	/* foo/. --> foo/ */
+	if (f[1] == 0)
+	  {
+	     *f = 0;
+	     continue;
+	  }
+
+	if (f[1] != '.')
+	  {
+	     f++;
+	     continue;
+	  }
+
+	f1 = f;
+	if (f[2] == SLASH_CHAR)
+	  f1 = f + 3;		       /* /../ */
+	else if (f[2] == 0)
+	  f1 = f + 2;		       /* /.. */
+	else
+	  continue;
+
+	/* /foo/blah/../bar. 
+	 * Move bar to /blah 
+	 */
+	f--;
+	if (f != file)
+	  f--;
+	while ((f != file) && (*f != SLASH_CHAR))
+	  f--;
+	str_move_bytes (f+1, f1);
+     }
+   return 0;
+}
+
+/* This function may free its argument and returned a new malloced copy */
+static char *make_abs_filename (char *file)
+{
+# if defined(IBMPC_SYSTEM)
+   char buf[4];
+# endif
+   char *dir, *file1;
+
+   if (*file != SLASH_CHAR)
+     {
+# if defined(IBMPC_SYSTEM)
+	if (*file && (file[1] == ':'))
+	  {
+	     if (file[2] == SLASH_CHAR)
+	       return file;
+	     /* Insert a slash */
+	     buf[0] = file[0]; buf[1] = ':'; buf[2] = SLASH_CHAR; buf[3] = 0;
+	     file1 = strcat_malloc (buf, file+2);
+	     SLfree (file);
+	     return file1;
+	  }
+   
+# endif
+	/* Otherwise it looks like a relative path.  */
+	/* get_cwd returns an absolute path with the trailing /. */
+	dir = jed_get_cwd ();
+	file1 = strcat_malloc (dir, file);
+	SLfree (file);
+	if (file1 == NULL)
+	  return NULL;
+	file = file1;
+     }
+
+# if defined(IBMPC_SYSTEM)
+   /* If file is not //share/foo, then prepend drive specifier */
+   if (file[0] == SLASH_CHAR)
+     {
+	if (file[1] != SLASH_CHAR)
+	  {	     
+	     buf[0] = 'A' + (pcsystem_getdisk ()-1); buf[1] = ':'; buf[2] = 0;
+	     file1 = strcat_malloc (buf, file);
+	     SLfree (file);
+	     return file1;
+	  }
+	/* Make sure the //share ends with a slash. */
+	file1 = file + 2;
+	while (*file1 && (*file1 != SLASH_CHAR))
+	  file1++;
+	if (*file1 != SLASH_CHAR)
+	  {
+	     buf[0] = SLASH_CHAR; buf[1] = 0;
+	     file1 = strcat_malloc (file, buf);
+	     SLfree (file);
+	     return file1;
+	  }
+     }
+# endif
+   return file;
+}
+
+static char *skip_drive_specifier (char *file)
+{
+#ifdef __QNX__
+   /* QNX pathnames look like //<node number>/rest_of_path 
+    *  -- we need to leave on the <node number> and return /rest_of_path
+    */
+   if ((file[0] == SLASH_CHAR) && (file[1] == SLASH_CHAR)
+       && (file[2] >= '1') && (file[2] <= '9'))
+     {
+	file += 2;
+	while (*file && (*file != SLASH_CHAR))
+	  file++;
+     }
+   return file;
+#endif			       /* __QNX__ */
+
+#ifdef IBMPC_SYSTEM
+   if (((file[0] != 0) && (file[1] == ':'))
+       || ((file[0] == SLASH_CHAR) && (file[1] == SLASH_CHAR)))
+     {	
+	file += 2;
+	while (*file && (*file != SLASH_CHAR))
+	  file++;
+     }
+#endif
+
+   return file;
+}
+
+static char *prep_filename (SLFUTURE_CONST char *filename)
+{
+   unsigned int len;
+   char *file;
+
+   len = strlen (filename);
+
+#ifdef MSWINDOWS
+   /* Strip quotes from around the filename, if any */
+   if ((len >= 2) && (filename[0] == '"') && (filename[len - 1] == '"'))
+     {
+	len -= 2;
+	filename++;
+     }
+#endif
+
+   if (NULL == (file = SLmake_nstring (filename, len)))
+     return NULL;
+
+#ifndef __unix__
+   slash2slash (file);
+#endif
+   return file;
+}
+
+char *jed_standardize_filename (SLFUTURE_CONST char *filename)
+{
+   char *file1, *file;
+   
+   if (NULL == (file = prep_filename (filename)))
+     return NULL;
+
+   if (NULL == (file = make_abs_filename (file)))
+     return NULL;
+   
+   file1 = skip_drive_specifier (file);
+
+   (void) remove_slash_slash (file1);
+
+   (void) remove_dot_dot_slashes (file1);
+   
+   return file;
+}
+
+   
+/* this routine returns a Static pointer which is considered volatile */
+char *jed_standardize_filename_static (char *file) /*{{{*/
+{
+   static char filebuf [JED_MAX_PATH_LEN];
+
+   *filebuf = 0;
+   if (NULL != (file = jed_standardize_filename (file)))
+     {
+	safe_strcpy (filebuf, file, JED_MAX_PATH_LEN);
+	SLfree (file);
+     }
+   return filebuf;
+}
 /*}}}*/
+
 #endif /* ! VMS */
 
 #ifdef sequent
@@ -589,7 +780,7 @@ void fixup_dir(char *dir) /*{{{*/
 #ifndef VMS
    int n;
    
-   if ((n = strlen(dir)) > 1)
+   if ((n = strlen(dir)) >= 1)
      {
 	n--;
 # if defined(IBMPC_SYSTEM)
@@ -644,22 +835,6 @@ int make_random_number (int *seed, int *max) /*{{{*/
 
 /*}}}*/
 
-#ifndef __GO32__
-# ifdef __unix__
-/* if non-zero, Flow control is enabled */
-void enable_flow_control(int *mode) /*{{{*/
-{
-   /* This kills X windows.  For the time being, work around it as follows */
-   if (X_Init_Term_Hook != NULL) return;
-   Flow_Control = *mode;
-   reset_tty();
-   init_tty();
-}
-
-/*}}}*/
-
-# endif
-#endif
 
 #if defined(IBMPC_SYSTEM)
 /* This routine converts  C:\  --> C:\ and C:\subdir\  -> C:\subdir */
@@ -776,3 +951,176 @@ void jed_pause (int *ms) /*{{{*/
 }
 
 /*}}}*/
+
+
+
+int jed_handle_interrupt (void)
+{
+#if SLANG_VERSION >= 20000
+   return SLang_handle_interrupt ();
+#else
+   return 0;
+#endif
+}
+
+void jed_sleep (unsigned int n)
+{
+   sleep (n);
+}
+
+
+char *jed_get_cwd (void) /*{{{*/
+{
+   static char cwd[JED_MAX_PATH_LEN];
+   char *c;
+
+#ifdef HAVE_GETCWD
+# if defined (__EMX__)
+   c = _getcwd2(cwd, sizeof(cwd));		       /* includes drive specifier */
+# else
+   c = getcwd(cwd, sizeof(cwd));		       /* djggp includes drive specifier */
+# endif
+#else
+   c = (char *) getwd(cwd);
+#endif
+   
+   if (c == NULL)
+     {
+#ifndef REAL_UNIX_SYSTEM
+	exit_error ("Unable to get the current working directory.", 0);
+#else
+	struct stat st1, st2;
+	
+	if ((NULL == (c = getenv ("PWD")))
+	    || (-1 == stat (c, &st1))
+	    || (-1 == stat (".", &st2))
+	    || (st1.st_dev != st2.st_dev)
+	    || (st1.st_ino != st2.st_ino))
+	  {
+	     jed_vmessage (1, "Unable to get CWD.  Assuming /");
+	     if (-1 == chdir ("/"))
+	       exit_error ("Unable to change directory to /", 0);
+	     c = "/";
+	  }
+
+	strncpy (cwd, c, sizeof (cwd));
+	cwd [sizeof (cwd) - 1] = 0;
+#endif
+     }
+
+#ifndef VMS
+   slash2slash(cwd);
+   fixup_dir(cwd);
+#endif
+   return(cwd);
+}
+
+/*}}}*/
+
+#ifndef VMS
+
+static char *prepend_home_dir (char *file)
+{
+   char *home;
+
+   home = getenv ("HOME");
+   if ((home == NULL) || (*home == 0))
+     home = "/";
+   
+   file = strcat_malloc (home, file);
+   if (file == NULL)
+     return file;
+   
+   home = jed_standardize_filename (file);
+   SLfree (file);
+   return home;
+}
+
+
+/* This function makes the following transformations:
+ * 
+ *     /foo/bar//root/baz   --> /root/baz
+ *     /foo/bar/~/foo       --> $HOME/foo
+ * 
+ * and then converts the result to standard form.
+ * Possible cases include:
+ * 
+ *    /foo/bar//tmp/x.c    --> /tmp/x.c
+ *    /foo/bar/~/me/baz    --> $HOME/me/baz
+ *    /foo/bar//tmp/~/bar  --> /tmp/~/bar
+ *    ~/bar                --> $HOME/bar
+ *    /tmp/~/bar//home/me  --> /home/me
+ * 
+ * DOS/Windows:
+ * 
+ *    C:/foo/bar/D:/foo     --> D:/foo
+ *    //foo/bar///share/baz --> //share/baz
+ * 
+ * For the case: /tmp/~/bar/~/foo, assume ~/foo.
+ */
+char *jed_expand_filename (char *file)
+{
+   char *f;
+   char *drive;
+
+   if (NULL == (file = prep_filename (file)))
+     return NULL;
+
+   drive = NULL;
+   f = skip_drive_specifier (file);
+
+   /* Now look for // */
+   while (1)
+     {
+	char *f1;
+
+	while (*f && (*f != SLASH_CHAR))
+	  f++;
+	if (*f == 0)
+	  break;
+	f++;
+	if (*f != SLASH_CHAR)
+	  {
+	     f1 = skip_drive_specifier (f);
+	     if (f1 == f)
+	       continue;
+	  }
+	drive = f;
+	break;
+     }
+   
+   if (drive != NULL)
+     {
+	f = jed_standardize_filename (drive);
+	SLfree (file);
+	return f;
+     }
+   
+   /* Now look for /~/ */
+   f = file + strlen (file);
+   while (f != file)
+     {
+	while (f > file)
+	  {
+	     f--;
+	     if (*f == '~')
+	       break;
+	  }
+	if (*f != '~')
+	  break;
+	if (f[1] != SLASH_CHAR)
+	  continue;
+	if ((f != file) && (*(f-1) != SLASH_CHAR))
+	  continue;
+	
+	f = prepend_home_dir (f + 1);
+	SLfree (file);
+	return f;
+     }
+
+   f = jed_standardize_filename (file);
+   SLfree (file);
+   return f;
+}
+
+#endif				       /* !VMS */
