@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -200,7 +200,7 @@ static int dos_system(char *command_line)
    int nCmdShow = SW_SHOWMINIMIZED;
    int wait = TRUE;
 
-   _SLsnprintf(buf, sizeof(buf), "%s\\bin\\mswshell.pif %s", Jed_Root_Dir, command_line);
+   SLsnprintf(buf, sizeof(buf), "%s\\bin\\mswshell.pif %s", Jed_Root_Dir, command_line);
 
    sys_delete_file(tmp_file);
 
@@ -293,7 +293,7 @@ int sys_findfirst(char *thefile)
 
    File_Attr = READON | SUBDIR;
 
-   file = expand_filename(thefile);
+   file = jed_standardize_filename_static(thefile);
    f1 = f = extract_file(file);
    strcpy (Found_Dir, file);
 
@@ -336,7 +336,7 @@ int sys_findnext(char *file)
 
 
 
-int sys_chmod(char *file, int what, int *mode, short *dum1, short *dum2)
+int sys_chmod(SLFUTURE_CONST char *file, int what, mode_t *mode, uid_t *dum1, gid_t *dum2)
 {
    int flag, m = *mode;
    (void) dum1; (void) dum2;
@@ -437,7 +437,7 @@ static int x_insert_cutbuffer (void)
 
 	     dat[x] = 0;
 
-	     jed_ins_chars((unsigned char *) dat, nbytes);
+	     jed_insert_nbytes((unsigned char *) dat, nbytes);
 	     SLfree (dat);
 	  }
 	GlobalUnlock(hBuf);
